import os

os.environ["CUDA_VISIBLE_DEVICES"] = "0"
os.environ["WANDB_MODE"] = "disabled"

import pytest
import torch
from datasets import Dataset
from transformers import AutoConfig, AutoModelForCausalLM, Trainer, TrainingArguments
from trl import SFTConfig, SFTTrainer

from bergson.data import load_gradients
from bergson.huggingface import (
    GradientCollectorCallback,
    prepare_for_gradient_collection,
)
from bergson.utils.utils import assert_type


class TestGradientCollectorCallback:
    @pytest.fixture
    def model(self):
        """Create a small test model."""
        config = AutoConfig.from_pretrained("trl-internal-testing/tiny-Phi3ForCausalLM")
        return AutoModelForCausalLM.from_config(config)

    @pytest.fixture
    def dataset(self):
        """Create a small test dataset."""
        data = {
            "input_ids": [
                [1, 2, 3, 4, 5],
                [6, 7, 8, 9, 10],
                [11, 12, 13, 14, 15],
                [16, 17, 18, 19, 20],
            ],
            "labels": [
                [1, 2, 3, 4, 5],
                [6, 7, 8, 9, 10],
                [11, 12, 13, 14, 15],
                [16, 17, 18, 19, 20],
            ],
        }
        return Dataset.from_dict(data)

    @pytest.mark.skipif(not torch.cuda.is_available(), reason="CUDA not available")
    def test_single_gpu_order_tracking(self, tmp_path, model, dataset):
        """Test that every step has an associated order record in single-GPU mode."""
        # Train the model with the callback
        training_args = TrainingArguments(
            output_dir=str(tmp_path / "output"),
            num_train_epochs=1,
            per_device_train_batch_size=2,
            per_device_eval_batch_size=2,
            gradient_accumulation_steps=1,
            save_strategy="no",
            logging_strategy="no",
            remove_unused_columns=False,
        )

        callback = GradientCollectorCallback(
            path=tmp_path / "gradients",
            track_order=True,
            use_optimizer_state=False,
        )

        trainer = Trainer(
            model=model,
            args=training_args,
            train_dataset=dataset,
            eval_dataset=dataset,
            callbacks=[callback],
        )
        trainer = prepare_for_gradient_collection(trainer)
        trainer.train()

        # Verify order records were created
        assert callback.order is not None
        assert len(callback.order) > 0

        # Check that every step has associated order records
        steps_with_records = set()
        for record in callback.order:
            steps_with_records.add(record["global_step"])

        # Get the actual number of training steps
        expected_steps = len(dataset) // training_args.per_device_train_batch_size
        if len(dataset) % training_args.per_device_train_batch_size != 0:
            expected_steps += 1

        # Verify we have records for all expected steps
        assert len(steps_with_records) == expected_steps
        # Expected steps are 1-indexed
        assert steps_with_records == set(range(1, expected_steps + 1))

        # Verify each record has required fields
        for record in callback.order:
            assert "_idx" in record
            assert "global_step" in record
            assert "epoch" in record
            assert isinstance(record["_idx"], int)
            assert isinstance(record["global_step"], int)
            assert isinstance(record["epoch"], int)

        # Verify indices are within valid range
        for record in callback.order:
            assert 0 <= record["_idx"] < len(dataset)

    @pytest.mark.skipif(not torch.cuda.is_available(), reason="CUDA not available")
    def test_order_tracking_disabled(self, tmp_path, model, dataset):
        """Test that no order records are created when tracking is disabled."""
        # Train the model with the callback
        training_args = TrainingArguments(
            output_dir=str(tmp_path / "output"),
            num_train_epochs=1,
            per_device_train_batch_size=2,
            per_device_eval_batch_size=2,
            gradient_accumulation_steps=1,
            save_strategy="no",
            logging_strategy="no",
            remove_unused_columns=False,
        )

        callback = GradientCollectorCallback(
            path=tmp_path / "gradients", use_optimizer_state=False
        )

        trainer = Trainer(
            model=model,
            args=training_args,
            train_dataset=dataset,
            eval_dataset=dataset,
            callbacks=[callback],
        )
        trainer = prepare_for_gradient_collection(trainer)
        trainer.train()

        # Verify no order records were created
        assert callback.order is None

    @pytest.mark.skipif(not torch.cuda.is_available(), reason="CUDA not available")
    def test_order_save_and_load(self, tmp_path, model, dataset):
        """Test that order records are properly saved and can be loaded."""
        # Train the model with the callback
        training_args = TrainingArguments(
            output_dir=str(tmp_path / "output"),
            num_train_epochs=1,
            per_device_train_batch_size=2,
            per_device_eval_batch_size=2,
            gradient_accumulation_steps=1,
            save_strategy="no",
            logging_strategy="no",
            remove_unused_columns=False,
        )

        callback = GradientCollectorCallback(
            path=tmp_path / "gradients",
            track_order=True,
            use_optimizer_state=False,
        )

        trainer = Trainer(
            model=model,
            args=training_args,
            train_dataset=dataset,
            eval_dataset=dataset,
            callbacks=[callback],
        )
        trainer = prepare_for_gradient_collection(trainer)
        trainer.train()

        # Verify order records were created
        assert callback.order is not None
        assert len(callback.order) > 0

        # Check that order file was saved
        order_file = tmp_path / "gradients" / "order.hf"
        assert order_file.exists()

        # Load and verify the saved order
        saved_order = Dataset.load_from_disk(str(order_file))
        assert len(saved_order) == len(callback.order)

        # Verify the saved order matches the in-memory order
        for i, record in enumerate(saved_order):
            record = assert_type(dict, record)
            assert record["_idx"] == callback.order[i]["_idx"]
            assert record["global_step"] == callback.order[i]["global_step"]
            assert record["epoch"] == callback.order[i]["epoch"]

    @pytest.mark.skipif(not torch.cuda.is_available(), reason="CUDA not available")
    def test_sft_trainer(self, tmp_path, model, dataset):
        """Test that gradient and order files are created and
        can be loaded after training with SFTTrainer."""
        # Set up and train the model with SFT
        sft_config = SFTConfig(
            output_dir=str(tmp_path / "sft_output"),
            num_train_epochs=1,
            per_device_train_batch_size=2,
            per_device_eval_batch_size=2,
            gradient_accumulation_steps=1,
            save_strategy="no",
            logging_strategy="no",
            remove_unused_columns=False,
        )

        callback = GradientCollectorCallback(
            path=tmp_path / "gradients",
            track_order=True,
            use_optimizer_state=False,
        )

        trainer = SFTTrainer(
            model=model,
            args=sft_config,
            train_dataset=dataset,
            eval_dataset=dataset,
            callbacks=[callback],
        )
        trainer = prepare_for_gradient_collection(trainer)
        trainer.train()

        # Verify training order was tracked
        assert callback.order is not None
        assert len(callback.order) > 0

        # Verify gradient files were created
        gradient_dir = tmp_path / "gradients"
        train_gradient_dir = gradient_dir / "train" / "epoch_0"

        assert gradient_dir.exists()
        assert train_gradient_dir.exists()
        assert (train_gradient_dir / "gradients.bin").exists()
        assert (train_gradient_dir / "info.json").exists()
        assert (gradient_dir / "order.hf").exists()

        # Test loading the gradient data directly
        gradients = load_gradients(train_gradient_dir)
        assert len(gradients) > 0

        # Verify order data was saved and can be loaded
        order_file = tmp_path / "gradients" / "order.hf"
        assert order_file.exists()

        saved_order = Dataset.load_from_disk(str(order_file))
        assert len(saved_order) > 0
        assert all(key in saved_order[0] for key in ["_idx", "global_step", "epoch"])



import logging
import sys
from typing import Optional

import torch.distributed as dist

# Set up logging
logging.basicConfig(
    level=logging.INFO,
    format="%(filename)s:%(funcName)s:%(lineno)d:%(levelname)s:  %(message)s",
    stream=sys.stdout,  # Ensure output goes to stdout
    force=True,  # Override any existing configuration
)


class RankFilter(logging.Filter):
    """Filter that only allows logging from rank 0."""

    def filter(self, record):
        rank = dist.get_rank() if dist.is_initialized() else 0
        return rank == 0


# Create a function to get loggers with consistent naming
def get_logger(
    name: Optional[str] = None,
    level: Optional[str] = None,
    disable_filter: bool = False,
) -> logging.Logger:
    """
    Get a logger with the configured format.

    Args:
        name: Logger name. If None, uses the calling module's __name__
        level: Logging level ("DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL").
               If None, uses the global default.

    Returns:
        Configured logger instance
    """
    if name is None:
        # Get the calling module's name
        import inspect

        frame = inspect.currentframe().f_back  # type: ignore
        assert frame is not None, "Frame should not be None"
        name = frame.f_globals.get("__name__", "unknown")

    logger = logging.getLogger(name)

    # Add rank filter if not already present
    if not disable_filter and not any(
        isinstance(f, RankFilter) for f in logger.filters
    ):
        logger.addFilter(RankFilter())

    if level is not None:
        # Convert string level to logging constant
        level_mapping = {
            "DEBUG": logging.DEBUG,
            "INFO": logging.INFO,
            "WARNING": logging.WARNING,
            "ERROR": logging.ERROR,
            "CRITICAL": logging.CRITICAL,
        }
        log_level = level_mapping.get(level.upper(), logging.INFO)
        logger.setLevel(log_level)

    return logger



import json
import math
import os
import shutil
from dataclasses import asdict
from datetime import timedelta

import torch
import torch.distributed as dist
from datasets import Dataset, IterableDataset
from tqdm.auto import tqdm

from bergson.collection import collect_gradients
from bergson.collector.gradient_collectors import GradientCollector
from bergson.config import IndexConfig, ReduceConfig
from bergson.data import allocate_batches
from bergson.utils.utils import assert_type
from bergson.utils.worker_utils import setup_model_and_peft

from .distributed import launch_distributed_run
from .utils.worker_utils import create_processor, setup_data_pipeline


def reduce_worker(
    rank: int,
    local_rank: int,
    world_size: int,
    index_cfg: IndexConfig,
    reduce_cfg: ReduceConfig,
    ds: Dataset | IterableDataset,
):
    """
    Distributed worker that aggregates per-document gradients into a single vector.

    Parameters
    ----------
    rank : int
        Distributed rank / GPU ID for this worker.
    local_rank : int
        Local rank / GPU ID for this worker on the node.
    world_size : int
        Total number of workers participating in the run.
    index_cfg : IndexConfig
        Specifies the model, tokenizer, PEFT adapters, and other settings.
    reduce_cfg : ReduceConfig
        Specifies aggregation strategy (mean/sum, unit normalization).
    ds : Dataset | IterableDataset
        The entire dataset to be indexed. A subset is assigned to each worker.
    """
    torch.cuda.set_device(local_rank)

    # These should be set by the main process
    if world_size > 1:
        addr = os.environ.get("MASTER_ADDR", "localhost")
        port = os.environ.get("MASTER_PORT", "29500")

        dist.init_process_group(
            "nccl",
            init_method=f"tcp://{addr}:{port}",
            device_id=torch.device(f"cuda:{local_rank}"),
            rank=rank,
            timeout=timedelta(minutes=30),
            world_size=world_size,
        )

    model, target_modules = setup_model_and_peft(index_cfg)
    processor = create_processor(model, ds, index_cfg, target_modules)

    attention_cfgs = {
        module: index_cfg.attention for module in index_cfg.split_attention_modules
    }

    kwargs = {
        "model": model,
        "data": ds,
        "processor": processor,
        "cfg": index_cfg,
        "target_modules": target_modules,
        "attention_cfgs": attention_cfgs,
        "reduce_cfg": reduce_cfg,
    }

    if isinstance(ds, Dataset):
        batches = allocate_batches(ds["length"], index_cfg.token_batch_size)
        kwargs["batches"] = batches
        collect_gradients(**kwargs)
    else:
        # Convert each shard to a Dataset then map over its gradients
        buf, shard_id = [], 0

        def flush(kwargs):
            nonlocal buf, shard_id
            if not buf:
                return
            ds_shard = assert_type(Dataset, Dataset.from_list(buf))
            batches = allocate_batches(
                ds_shard["length"][:], index_cfg.token_batch_size
            )
            kwargs["ds"] = ds_shard
            kwargs["batches"] = batches
            collect_gradients(**kwargs)

            buf.clear()
            shard_id += 1

        for ex in tqdm(ds, desc="Collecting gradients"):
            buf.append(ex)
            if len(buf) == index_cfg.stream_shard_size:
                flush(kwargs=kwargs)

        flush(kwargs=kwargs)  # Final flush
        if rank == 0:
            processor.save(index_cfg.partial_run_path)

    # Save info.json for score command (if it doesn't already exist from Builder)
    if rank == 0:
        info_path = index_cfg.partial_run_path / "info.json"
        if not info_path.exists():
            # Create temporary collector to get shapes
            shapes = GradientCollector(
                model=model,
                data=Dataset.from_list([]),
                processor=processor,
                cfg=index_cfg,
                target_modules=target_modules,
            ).shapes()

            grad_sizes = {name: math.prod(s) for name, s in shapes.items()}

            # Build dtype structure matching create_index format for consistency
            struct_dtype = {
                "names": list(grad_sizes.keys()),
                "formats": [f"({size},)<f4" for size in grad_sizes.values()],
                "itemsize": 4 * sum(grad_sizes.values()),
            }

            metadata = {
                "num_grads": 1,
                "dtype": struct_dtype,
                "grad_sizes": grad_sizes,
                "base_dtype": "float32",
            }

            with info_path.open("w") as f:
                json.dump(metadata, f, indent=2)


def reduce(index_cfg: IndexConfig, reduce_cfg: ReduceConfig):
    """
    Reduce a dataset to a single aggregated gradient vector.

    Parameters
    ----------
    index_cfg : IndexConfig
        Specifies the run path, dataset, model, tokenizer, PEFT adapters,
        and many other gradient collection settings.
    reduce_cfg : ReduceConfig
        Specifies aggregation strategy (mean/sum, unit normalization).
    """
    index_cfg.partial_run_path.mkdir(parents=True, exist_ok=True)
    with (index_cfg.partial_run_path / "index_config.json").open("w") as f:
        json.dump(asdict(index_cfg), f, indent=2)

    ds = setup_data_pipeline(index_cfg)

    launch_distributed_run(
        "reduce", reduce_worker, [index_cfg, reduce_cfg, ds], index_cfg.distributed
    )

    if index_cfg.distributed.rank == 0:
        shutil.move(index_cfg.partial_run_path, index_cfg.run_path)



import json
import os
import shutil
from dataclasses import asdict
from datetime import timedelta

import torch
import torch.distributed as dist
from datasets import Dataset, IterableDataset
from tqdm.auto import tqdm

from bergson.collection import collect_gradients
from bergson.config import IndexConfig
from bergson.data import allocate_batches
from bergson.distributed import launch_distributed_run
from bergson.utils.utils import assert_type, setup_reproducibility
from bergson.utils.worker_utils import (
    create_processor,
    setup_data_pipeline,
    setup_model_and_peft,
)


def build_worker(
    rank: int,
    local_rank: int,
    world_size: int,
    cfg: IndexConfig,
    ds: Dataset | IterableDataset,
):
    """
    Build worker executed per rank to collect gradients to populate the index.

    Parameters
    ----------
    rank : int
        Distributed rank / GPU ID for this worker.
    local_rank : int
        Local rank / GPU ID for this worker on the node.
    world_size : int
        Total number of workers participating in the run.
    cfg : IndexConfig
        Specifies the model, tokenizer, PEFT adapters, and other settings.
    ds : Dataset | IterableDataset
        The entire dataset to be indexed. A subset is assigned to each worker.
    """
    torch.cuda.set_device(local_rank)

    # These should be set by the main process
    if world_size > 1:
        addr = os.environ.get("MASTER_ADDR", "localhost")
        port = os.environ.get("MASTER_PORT", "29500")

        dist.init_process_group(
            "nccl",
            init_method=f"tcp://{addr}:{port}",
            device_id=torch.device(f"cuda:{local_rank}"),
            rank=rank,
            timeout=timedelta(minutes=2),
            world_size=world_size,
        )

    model, target_modules = setup_model_and_peft(cfg)
    processor = create_processor(model, ds, cfg, target_modules)

    attention_cfgs = {module: cfg.attention for module in cfg.split_attention_modules}

    kwargs = {
        "model": model,
        "data": ds,
        "processor": processor,
        "cfg": cfg,
        "target_modules": target_modules,
        "attention_cfgs": attention_cfgs,
    }

    if isinstance(ds, Dataset):
        batches = allocate_batches(ds["length"], cfg.token_batch_size)
        kwargs["batches"] = batches
        collect_gradients(**kwargs)
    else:
        # Convert each shard to a Dataset then map over its gradients
        buf, shard_id = [], 0

        def flush(kwargs):
            nonlocal buf, shard_id
            if not buf:
                return
            ds_shard = assert_type(Dataset, Dataset.from_list(buf))
            batches = allocate_batches(ds_shard["length"][:], cfg.token_batch_size)
            kwargs["ds"] = ds_shard
            kwargs["batches"] = batches
            collect_gradients(**kwargs)

            buf.clear()
            shard_id += 1

        for ex in tqdm(ds, desc="Collecting gradients"):
            buf.append(ex)
            if len(buf) == cfg.stream_shard_size:
                flush(kwargs=kwargs)

        flush(kwargs=kwargs)  # Final flush
        if rank == 0:
            processor.save(cfg.partial_run_path)


def build(index_cfg: IndexConfig):
    """
    Build a gradient index by distributing work across all available GPUs.

    Parameters
    ----------
    index_cfg : IndexConfig
        Specifies the run path, dataset, model, tokenizer, PEFT adapters,
        and many other gradient collection settings.
    """
    if index_cfg.debug:
        setup_reproducibility()

    index_cfg.partial_run_path.mkdir(parents=True, exist_ok=True)
    with (index_cfg.partial_run_path / "index_config.json").open("w") as f:
        json.dump(asdict(index_cfg), f, indent=2)

    ds = setup_data_pipeline(index_cfg)

    launch_distributed_run(
        "build", build_worker, [index_cfg, ds], index_cfg.distributed
    )

    rank = index_cfg.distributed.rank
    if rank == 0:
        shutil.move(index_cfg.partial_run_path, index_cfg.run_path)



import os
import socket
from typing import Any, Callable

import torch.distributed as dist
import torch.multiprocessing as mp
from torch.distributed.elastic.multiprocessing import DefaultLogsSpecs, start_processes

from .config import DistributedConfig


def dist_worker(
    worker: Callable,
    *worker_args,
):
    try:
        worker(*worker_args)
    finally:
        if dist.is_initialized():
            try:
                dist.barrier()
            except Exception as e:
                print(f"Barrier failed during cleanup: {e}")
                pass

            dist.destroy_process_group()


def launch_distributed_run(
    process_name: str,
    worker,
    const_worker_args: list[Any],
    dist_config: DistributedConfig | None = None,
):
    if dist_config is None:
        dist_config = DistributedConfig()

    local_world_size = dist_config.nproc_per_node
    world_size = dist_config.world_size
    start_rank = dist_config.start_rank

    # Multi-node environment
    if dist_config.nnode > 1:
        master_addr = os.environ.get("MASTER_ADDR", "localhost")
        master_port = os.environ.get("MASTER_PORT", "29500")
    else:
        master_addr = "localhost"
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            s.bind(("", 0))
            _, master_port = s.getsockname()
        master_port = str(master_port)

    if world_size <= 1:
        worker(0, 0, 1, *const_worker_args)
    else:
        mp.set_sharing_strategy("file_system")

        ctx = None
        try:
            ctx = start_processes(
                process_name,
                dist_worker,
                args={
                    i: (worker, start_rank + i, i, world_size, *const_worker_args)
                    for i in range(local_world_size)
                },
                envs={
                    i: {
                        "LOCAL_RANK": str(i),
                        "RANK": str(start_rank + i),
                        "WORLD_SIZE": str(world_size),
                        "MASTER_ADDR": master_addr,
                        "MASTER_PORT": master_port,
                    }
                    for i in range(local_world_size)
                },
                logs_specs=DefaultLogsSpecs(),
            )
            result = ctx.wait()

            if result is not None and hasattr(result, "failures") and result.failures:
                newline = "\n"
                raise RuntimeError(
                    f"{process_name} failed with {len(result.failures)} process "
                    f"failure(s): {newline.join([str(f) for f in result.failures])}"
                )
        finally:
            if ctx is not None:
                ctx.close()  # Kill any processes that are still running



import torch.nn as nn

from bergson import fit_normalizers
from bergson.config import IndexConfig


def test_fit_normalizers_runs(tmp_path, model, dataset):
    target_modules = {
        name
        for name, module in model.base_model.named_modules()
        if isinstance(module, nn.Linear)
    }
    print("len dataset", len(dataset))
    print("target_modules", target_modules)

    dataset = dataset.repeat(10)

    normalizers = fit_normalizers(
        model,
        dataset,
        cfg=IndexConfig(
            run_path=str(tmp_path),
            skip_preconditioners=True,
            normalizer="adam",
        ),
        batches=[[idx] for idx in range(len(dataset))],
        target_modules=target_modules,
    )

    assert len(normalizers) == len(target_modules)



import pytest
import torch

from bergson.utils.math import optimal_linear_shrinkage


@pytest.mark.parametrize(
    "p,n",
    [
        # Test the n < p case
        (32, 16),
        (64, 32),
        (128, 64),
        # And the n > p case
        (16, 64),
        (32, 128),
        (64, 256),
    ],
)
def test_olse_shrinkage(p: int, n: int):
    # Use CUDA to speed things up if available
    device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
    prng = torch.Generator(device=device).manual_seed(42)

    # Number of population matrices to test
    N = 100

    # Generate a random covariance matrix
    A = torch.randn(N, 1, p, p, device=device, generator=prng)
    S_true = A @ A.mH / p
    torch.linalg.diagonal(S_true).add_(1e-3)

    # Generate random means
    mu = torch.randn(N, 1, 1, p, device=device, generator=prng)

    # Number of draws of empirical samples per matrix
    m = 100

    # Generate random Gaussian vectors with this covariance matrix
    scale_tril = torch.linalg.cholesky(S_true)
    X = torch.randn(N, m, n, p, device=device, generator=prng) @ scale_tril.mH + mu

    # Compute the sample covariance matrix
    mu_hat = X.mean(dim=-2, keepdim=True)
    X_centered = X - mu_hat
    S_hat = (X_centered.mH @ X_centered) / n

    # Apply shrinkage
    S_olse = optimal_linear_shrinkage(S_hat, n)

    # Compute the average squared Frobenius error across the different draws
    norm_naive = torch.square(S_hat - S_true).sum(dim=(-1, -2)).mean(1)
    norm_olse = torch.square(S_olse - S_true).sum(dim=(-1, -2)).mean(1)

    # The MSE should be lower for all the population matrices
    assert torch.all(norm_olse <= norm_naive)

    # Compute the sample second moment matrix
    S_hat = S_hat + mu_hat.mT @ mu_hat
    S_olse = S_olse + mu_hat.mT @ mu_hat
    S_true = A + mu.mT @ mu

    # Compute the average squared Frobenius error across the different draws
    norm_naive = torch.square(S_hat - S_true).sum(dim=(-1, -2)).mean(1)
    norm_olse = torch.square(S_olse - S_true).sum(dim=(-1, -2)).mean(1)

    # The MSE should be lower for all the population matrices
    assert torch.all(norm_olse <= norm_naive)



import json
import math
import subprocess
import sys
from pathlib import Path

import numpy as np
import pytest
import torch
from ml_dtypes import bfloat16
from transformers import AutoConfig, AutoModelForCausalLM

from bergson import (
    GradientProcessor,
    collect_gradients,
)
from bergson.collector.gradient_collectors import GradientCollector
from bergson.config import IndexConfig, ScoreConfig
from bergson.data import create_index, load_scores
from bergson.score.score import precondition_grads
from bergson.score.score_writer import MemmapScoreWriter
from bergson.score.scorer import Scorer
from bergson.utils.utils import (
    convert_precision_to_torch,
    get_gradient_dtype,
    tensor_to_numpy,
)


@pytest.mark.skipif(not torch.cuda.is_available(), reason="CUDA not available")
def test_large_gradients_query(tmp_path: Path, dataset):
    # Create index for uncompressed gradients from a large model.
    config = AutoConfig.from_pretrained(
        "EleutherAI/pythia-1.4b", trust_remote_code=True
    )
    model = AutoModelForCausalLM.from_config(config)

    collector = GradientCollector(
        model.base_model, data=dataset, cfg=IndexConfig(run_path=str(tmp_path))
    )
    grad_sizes = {name: math.prod(s) for name, s in collector.shapes().items()}

    dataset.save_to_disk(str(tmp_path / "query_ds" / "data.hf"))
    create_index(
        tmp_path / "query_ds",
        num_grads=len(dataset),
        grad_sizes=grad_sizes,
        dtype=np.float32,
        with_structure=False,
    )

    result = subprocess.run(
        [
            sys.executable,
            "-m",
            "bergson",
            "score",
            "test_score_e2e",
            "--projection_dim",
            "0",
            "--query_path",
            str(tmp_path / "query_ds"),
            "--model",
            "EleutherAI/pythia-1.4b",
            "--dataset",
            "NeelNanda/pile-10k",
            "--split",
            "train[:8]",
            "--truncation",
            "--token_batch_size",
            "256",
            "--skip_preconditioners",
        ],
        cwd=tmp_path,
        capture_output=True,
        text=True,
    )

    assert result.returncode == 0
    assert (
        "error" not in result.stderr.lower()
    ), f"Error found in stderr: {result.stderr}"


@pytest.mark.skipif(not torch.cuda.is_available(), reason="CUDA not available")
def test_score(tmp_path: Path, model, dataset):
    processor = GradientProcessor(projection_dim=16)
    collector = GradientCollector(
        model.base_model,
        data=dataset,
        cfg=IndexConfig(run_path=str(tmp_path)),
        processor=processor,
    )
    shapes = collector.shapes()

    cfg = IndexConfig(run_path=str(tmp_path), token_batch_size=1024)
    score_cfg = ScoreConfig(
        query_path=str(tmp_path / "query_gradient_ds"),
        modules=list(shapes.keys()),
        score="mean",
    )

    query_grads = {
        module: torch.randn(1, shape.numel()) for module, shape in shapes.items()
    }

    score_dtype = (
        convert_precision_to_torch(score_cfg.precision)
        if score_cfg.precision != "auto"
        else get_gradient_dtype(model)
    )

    score_writer = MemmapScoreWriter(tmp_path, len(dataset), 1, dtype=score_dtype)
    scorer = Scorer(
        query_grads=query_grads,
        modules=list(shapes.keys()),
        writer=score_writer,
        device=torch.device("cpu"),
        dtype=score_dtype,
    )

    collect_gradients(
        model=model,
        data=dataset,
        processor=processor,
        cfg=cfg,
        scorer=scorer,
    )

    assert (tmp_path / "info.json").exists()
    assert (tmp_path / "scores.bin").exists()

    with open(tmp_path / "info.json", "r") as f:
        info = json.load(f)

    scores = load_scores(tmp_path)

    assert len(scores) == len(dataset)

    assert info["num_scores"] == 1

    assert np.allclose(scores["score_0"], np.array([1.8334405, 0.3371131]))


@pytest.mark.skipif(not torch.cuda.is_available(), reason="CUDA not available")
def test_precondition_ds(tmp_path: Path, model, dataset):
    cfg = IndexConfig(run_path=str(tmp_path), token_batch_size=1024)

    preprocess_device = torch.device("cuda:0")

    # Populate and save preconditioners
    processor = GradientProcessor(projection_dim=16)
    collector = GradientCollector(
        model.base_model,
        data=dataset,
        cfg=cfg,
        processor=processor,
    )
    collect_gradients(
        model=model,
        data=dataset,
        processor=processor,
        cfg=cfg,
    )
    processor.save(tmp_path)

    # Produce query gradients dict
    query_grads = {
        module: torch.randn(1, shape.numel())
        for module, shape in collector.shapes().items()
    }

    # Produce preconditioned query gradients
    score_cfg = ScoreConfig(
        query_path=str(tmp_path / "query_gradient_ds"),
        modules=list(collector.shapes().keys()),
        score="mean",
        query_preconditioner_path=str(tmp_path),
    )

    preconditioned = precondition_grads(
        query_grads, score_cfg, score_cfg.modules, preprocess_device
    )

    # Produce query gradients without preconditioning
    score_cfg.query_preconditioner_path = None

    vanilla = precondition_grads(
        query_grads, score_cfg, score_cfg.modules, preprocess_device
    )

    # Compare the two
    for name in score_cfg.modules:
        assert not torch.allclose(preconditioned[name], vanilla[name])


def test_memmap_score_writer_bfloat16(tmp_path: Path):
    """Test that MemmapScoreWriter correctly writes and reads bfloat16 scores."""
    num_items = 10
    num_scores = 3

    writer = MemmapScoreWriter(tmp_path, num_items, num_scores, dtype=torch.bfloat16)

    # Create some test scores in bfloat16
    scores_batch1 = torch.tensor(
        [[1.5, 2.5, 3.5], [4.5, 5.5, 6.5]], dtype=torch.bfloat16
    )
    scores_batch2 = torch.tensor(
        [[7.5, 8.5, 9.5], [10.5, 11.5, 12.5], [13.5, 14.5, 15.5]],
        dtype=torch.bfloat16,
    )

    # Write scores
    writer([0, 1], scores_batch1)
    writer([5, 6, 7], scores_batch2)
    writer.flush()

    # Verify the files exist
    assert (tmp_path / "scores.bin").exists()
    assert (tmp_path / "info.json").exists()

    # Read back and verify
    with open(tmp_path / "info.json", "r") as f:
        info = json.load(f)

    assert info["num_items"] == num_items
    assert info["num_scores"] == num_scores
    assert "bfloat16" in info["dtype"]["formats"][0]

    # Check written flags
    assert writer.scores["written_0"][0]
    assert writer.scores["written_0"][1]
    assert not writer.scores["written_0"][2]  # Not written
    assert writer.scores["written_0"][5]
    assert writer.scores["written_0"][6]
    assert writer.scores["written_0"][7]

    # Check score values (convert back to compare)
    expected_batch1 = tensor_to_numpy(scores_batch1)
    expected_batch2 = tensor_to_numpy(scores_batch2)

    np.testing.assert_array_equal(
        writer.scores["score_0"][[0, 1]].view(bfloat16), expected_batch1[:, 0]
    )
    np.testing.assert_array_equal(
        writer.scores["score_1"][[0, 1]].view(bfloat16), expected_batch1[:, 1]
    )
    np.testing.assert_array_equal(
        writer.scores["score_2"][[0, 1]].view(bfloat16), expected_batch1[:, 2]
    )

    np.testing.assert_array_equal(
        writer.scores["score_0"][[5, 6, 7]].view(bfloat16), expected_batch2[:, 0]
    )


def test_memmap_score_writer_float32(tmp_path: Path):
    """Test that MemmapScoreWriter correctly writes float32 scores (baseline)."""
    num_items = 5
    num_scores = 2

    writer = MemmapScoreWriter(tmp_path, num_items, num_scores, dtype=torch.float32)

    scores = torch.tensor([[1.5, 2.5], [3.5, 4.5]], dtype=torch.float32)
    writer([0, 1], scores)
    writer.flush()

    # Verify values
    np.testing.assert_array_almost_equal(
        writer.scores["score_0"][[0, 1]], np.array([1.5, 3.5], dtype=np.float32)
    )
    np.testing.assert_array_almost_equal(
        writer.scores["score_1"][[0, 1]], np.array([2.5, 4.5], dtype=np.float32)
    )



import math
from dataclasses import dataclass, field

import torch
import torch.distributed as dist
import torch.nn as nn
from datasets import Dataset, Value
from jaxtyping import Float
from torch import Tensor

from bergson.collector.collector import HookCollectorBase
from bergson.config import IndexConfig, ReduceConfig
from bergson.data import Builder
from bergson.gradients import (
    AdafactorNormalizer,
    AdamNormalizer,
    LayerAdapter,
)
from bergson.process_preconditioners import process_preconditioners
from bergson.score.scorer import Scorer
from bergson.utils.utils import assert_type, get_gradient_dtype


@dataclass(kw_only=True)
class GradientCollectorWithDistributedPreconditioners(HookCollectorBase):
    """
    Collects per-sample gradients from model layers and writes them to disk.
    Preconditioners are distributed across nodes, and data from each node is
    distributed to each preconditioner at every step. This enables the computation
    of preconditioners that are too large to fit on a single device.

    - For each forward/backward hook, we compute the the gradient or a low-rank
    approximation via random projections, if cfg.projection_dim is set.
    - Supports also normalization via Adam or Adafactor normalizers.
    - Uses Builder for index construction and gradient saving.
    - Also supports Scorer for on-the-fly scoring of gradients.
    """

    data: Dataset
    """The dataset being processed."""

    cfg: IndexConfig
    """Configuration for gradient index."""

    mod_grads: dict = field(default_factory=dict)
    """Temporary storage for gradients during a batch, keyed by module name."""

    reduce_cfg: ReduceConfig | None = None
    """Configuration for in-run gradient reduction."""

    builder: Builder | None = None
    """Handles writing gradients to disk. Created in setup() if save_index is True."""

    scorer: Scorer | None = None
    """Optional scorer for computing scores instead of building an index."""

    def __init__(self, *args, **kwargs):
        self.data = assert_type(Dataset, kwargs["data"])
        self.cfg = assert_type(IndexConfig, kwargs["cfg"])

        self.reduce_cfg = kwargs.get("reduce_cfg", None)
        self.builder = kwargs.get("builder", None)
        self.scorer = kwargs.get("scorer", None)
        self.mod_grads = {}

        # Extract parent class arguments
        parent_kwargs = {
            k: v
            for k, v in kwargs.items()
            if k
            in {
                "model",
                "filter_modules",
                "target_modules",
                "processor",
                "attention_cfgs",
            }
        }
        parent_kwargs["filter_modules"] = self.cfg.filter_modules

        super().__init__(*args, **parent_kwargs)

    def setup(self) -> None:
        """
        Initialize collector state.

        Sets up a Builder for gradient storage if not using a Scorer.
        """
        assert isinstance(
            self.model.device, torch.device
        ), "Model device is not set correctly"
        if self.cfg.include_bias and self.processor.normalizers is not None:
            raise NotImplementedError(
                "Bias with normalizers not supported yet, "
                "consider disabling bias inclusion for now."
            )

        self.owned_modules: set[str] = set()
        self.module_to_rank: dict[str, int] = {}

        self.save_dtype = get_gradient_dtype(self.model)
        self.lo = torch.finfo(self.save_dtype).min
        self.hi = torch.finfo(self.save_dtype).max

        self.per_doc_losses = torch.full(
            (len(self.data),),
            device=self.model.device,
            dtype=torch.float32,
            fill_value=0.0,
        )

        # Compute whether we need to save the index
        self.save_index = self.scorer is None and not self.cfg.skip_index

        if self.save_index:
            grad_sizes = {name: math.prod(s) for name, s in self.shapes().items()}
            self.builder = Builder(
                self.cfg.partial_run_path,
                self.data,
                grad_sizes,
                self.save_dtype,
                self.reduce_cfg,
            )
        else:
            self.builder = None

        if dist.is_initialized():
            rank = dist.get_rank()
            num_devices = dist.get_world_size()
            available_modules = list(self.shapes().keys())

            num_modules = len(available_modules)
            base, remainder = divmod(num_modules, num_devices)

            assert base > 0, "Each rank must own at least one module"

            start_idx = rank * base + min(rank, remainder)
            end_idx = start_idx + base + (1 if rank < remainder else 0)
            self.owned_modules = set(available_modules[start_idx:end_idx])

            for i, module_name in enumerate(available_modules):
                # Inverse of the start_idx formula
                self.module_to_rank[module_name] = (
                    min(i // (base + 1), remainder - 1)
                    if i < remainder * (base + 1)
                    else remainder + (i - remainder * (base + 1)) // base
                )

            print(f"Rank {rank} owns {len(self.owned_modules)} modules")
        else:
            self.owned_modules = set(self.shapes().keys())

    def forward_hook(self, module: nn.Module, a: Float[Tensor, "N S I"]) -> None:
        """
        Cache activations for gradient computation with normalizer preprocessing
        and compress via random projection if configured.
        Stores result in module._inputs for use in backward_hook.
        """
        p = self.processor.projection_dim
        name = assert_type(str, module._name)
        i = getattr(module, LayerAdapter.in_attr(module))
        normalizer = self.processor.normalizers.get(name)

        if isinstance(normalizer, AdamNormalizer):
            module._inputs = a
            return
        if isinstance(normalizer, AdafactorNormalizer):
            a_factor = normalizer.col.add(1e-30)
            a_factor = a_factor.rsqrt()
            a = a * a_factor.type_as(a)  # [N, S, I] * [I] → [N, S, I]

        if module._has_bias:
            # Append ones to activation for bias term
            ones = torch.ones(a.size(0), a.size(1), 1, device=a.device, dtype=a.dtype)
            a = torch.cat([a, ones], dim=-1)
            i = i + 1
            setattr(module, LayerAdapter.in_attr(module), i)
        if p is not None:
            a_projection = self.projection(name, p, i, "right", a.device, a.dtype).T
            a = a @ a_projection  # type: ignore
        # set module._inputs to a
        module._inputs = a

    @HookCollectorBase.split_attention_heads
    def backward_hook(self, module: nn.Module, g: Float[Tensor, "N S O"]):
        """
        Compute per-sample gradient and store in mod_grads.

        Computes gradient as outer product g.T @ a (again with optional projection and
        normalization).
        """
        a = module._inputs  # [N, S, I/q]

        assert isinstance(a, torch.Tensor), "Activation cache missing for module"
        name = assert_type(str, module._name)
        p = self.processor.projection_dim
        i = getattr(module, LayerAdapter.in_attr(module))
        o = getattr(module, LayerAdapter.out_attr(module))
        normalizer = self.processor.normalizers.get(name)

        if isinstance(normalizer, AdamNormalizer):
            full_gradient = g.mT @ a  # [N, O, S] @ [N, S, I] → [N, O, I]
            P = normalizer.normalize_(full_gradient)
            if p is not None:
                g_projection = self.projection(name, p, o, "left", g.device, g.dtype)
                a_projection = self.projection(name, p, i, "right", g.device, g.dtype).T
                P = g_projection @ P @ a_projection
        else:
            if isinstance(normalizer, AdafactorNormalizer):
                g_factor = normalizer.row.add(1e-30)
                g_factor = g_factor.mean().sqrt() * g_factor.rsqrt()
                g = g * g_factor.type_as(g)  # [N, S, O] * [O] → [N, S, O]

            if p is not None:
                g_projection = self.projection(name, p, o, "left", g.device, g.dtype)
                g = g @ g_projection.T  # [N, S, p]

            P = g.mT @ a  # [N, O/p, S] @ [N, S, I/q] → [N, O/p, I/q]

        P = P.flatten(1).clamp_(self.lo, self.hi)

        # Keep gradients in original dtype for preconditioner computation
        self.mod_grads[name] = P

        if self.cfg.skip_preconditioners:
            if self.save_index:
                # Asynchronously move the gradient to CPU and convert to the final dtype
                self.mod_grads[name] = P.to(
                    device="cpu", dtype=self.save_dtype, non_blocking=True
                )
            else:
                self.mod_grads[name] = P.to(dtype=self.save_dtype)

        del module._inputs

    def process_batch(self, indices: list[int], **kwargs):
        """Process collected gradients for a batch and update losses."""
        losses = kwargs.get("losses")
        assert losses is not None, "losses must be provided in kwargs"

        # Send gradients to owning ranks and compute outer products there
        if not self.cfg.skip_preconditioners:
            exchange_preconditioner_gradients(
                self.mod_grads,
                self.processor.preconditioners,
                self.module_to_rank,
                self.owned_modules,
                self.rank,
            )

            # Convert mod_grads to the right dtype for save_index logic
            if self.save_index:
                for name in self.mod_grads:
                    self.mod_grads[name] = self.mod_grads[name].to(
                        device="cpu", dtype=self.save_dtype, non_blocking=True
                    )
            else:
                for name in self.mod_grads:
                    self.mod_grads[name] = self.mod_grads[name].to(
                        dtype=self.save_dtype
                    )

        if self.builder:
            self.builder(indices, self.mod_grads)
        if self.scorer:
            self.scorer(indices, self.mod_grads)
        self.mod_grads.clear()
        self.per_doc_losses[indices] = losses.detach().type_as(self.per_doc_losses)

    def teardown(self):
        """
        Finalize gradient collection, save results and flush/reduce the Builder.
        """
        assert isinstance(
            self.cfg, IndexConfig
        ), "cfg is required for GradientCollector"  # pleasing type checker
        if dist.is_initialized():
            dist.reduce(self.per_doc_losses, dst=0)

        grad_sizes = {name: math.prod(s) for name, s in self.shapes().items()}
        if self.processor.preconditioners:
            process_preconditioners(
                self.processor,
                self.processor.preconditioners,
                len(self.data),
                grad_sizes,
                self.rank,
            )

        if self.rank == 0:
            if self.cfg.drop_columns:
                self.data = self.data.remove_columns(["input_ids"])

            self.data = self.data.add_column(
                "loss",
                self.per_doc_losses.cpu().numpy(),
                feature=Value("float32"),
                new_fingerprint="loss",
            )

            self.data.save_to_disk(self.cfg.partial_run_path / "data.hf")

            self.processor.save(self.cfg.partial_run_path)

        # Flush and reduce builder if it exists
        if self.builder is not None:
            self.builder.flush()
            self.builder.dist_reduce()


def exchange_preconditioner_gradients(
    mod_grads: dict[str, torch.Tensor],
    preconditioners: dict[str, torch.Tensor],
    module_to_rank: dict[str, int],
    owned_modules: set[str],
    rank: int,
):
    """
    Send gradients to the ranks that own their preconditioners, and accumulate
    outer products on the owning ranks.
    Each rank sends gradients for modules it doesn't own to the owning ranks,
    and receives gradients for modules it owns to compute outer products.
    """
    # Process current rank data for owned modules
    for name, g in mod_grads.items():
        if name not in owned_modules:
            continue

        g = g.float()
        if name in preconditioners:
            preconditioners[name].addmm_(g.mT, g)
        else:
            preconditioners[name] = g.mT @ g

    if not dist.is_initialized():
        return

    world_size = dist.get_world_size()
    device = next(iter(mod_grads.values())).device

    module_names = list(mod_grads.keys())
    module_numel = {n: int(mod_grads[n].numel()) for n in module_names}

    current_rank_chunk = torch.empty(0, device=device, dtype=torch.float32)

    # Flatten batch dimension: all to all works on contiguous 1-D tensors
    send_chunks = [
        (
            current_rank_chunk
            if dest == rank
            else torch.cat(
                [
                    mod_grads[name].flatten()
                    for name in module_names
                    if module_to_rank[name] == dest
                ]
            )
        )
        for dest in range(world_size)
    ]

    # --- collective exchange of gradient sizes in order of mod_grads ---
    send_sizes = torch.tensor(
        [t.numel() for t in send_chunks], device=device, dtype=torch.int64
    )
    recv_sizes = torch.empty_like(send_sizes)

    dist.all_to_all_single(recv_sizes, send_sizes)

    # --- collective exchange of gradient in order of mod_grads ---
    send_buf = torch.cat(send_chunks)
    recv_buf = torch.empty(
        int(recv_sizes.sum().item()), device=device, dtype=torch.float32
    )

    dist.all_to_all_single(
        recv_buf,
        send_buf,
        output_split_sizes=recv_sizes.tolist(),
        input_split_sizes=send_sizes.tolist(),
    )

    # Unpack gradients in src-rank order
    # Within each src partition, modules are in fixed order.
    offset = 0
    for src_rank in range(world_size):
        part_len = int(recv_sizes[src_rank].item())
        part = recv_buf[offset : offset + part_len]
        offset += part_len

        if part_len == 0 or src_rank == rank:
            continue

        p = 0
        for name in owned_modules:
            n = module_numel[name]
            flat = part[p : p + n]
            p += n

            feature_dim = mod_grads[name].shape[-1]
            g = flat.to(device, non_blocking=True).view(-1, feature_dim).float()

            if name in preconditioners:
                preconditioners[name].addmm_(g.mT, g)
            else:
                preconditioners[name] = g.mT @ g



from peft import PeftModel, get_peft_model_state_dict
from transformers import PreTrainedModel


def detect_peft_modules(model: PeftModel) -> set[str]:
    """
    Infer the names of PEFT modules from a wrapped model.

    Returns
    -------
    set[str]
        Fully-qualified module names that correspond to adapter weights.
    """
    # Extract target modules
    target_modules = set()
    peft_state_dict = get_peft_model_state_dict(model)

    for adapter_name in model.peft_config:
        for name in peft_state_dict:
            prefix = name.removesuffix(".weight")
            processed_name = f"{prefix}.{adapter_name}".removeprefix("base_model.")
            try:
                model.get_submodule(processed_name)
                target_modules.add(processed_name)
            except AttributeError as e:
                raise RuntimeError(
                    f"Adapter parameter '{processed_name}' not found in the model."
                ) from e

    return target_modules


def set_peft_enabled(model: PreTrainedModel, enabled: bool):
    """Bizarrely, the standard enable_adapters and disable_adapters aren't working"""
    from peft.tuners.tuners_utils import BaseTunerLayer
    from peft.utils import ModulesToSaveWrapper

    futile = True

    for _, module in model.named_modules():
        if isinstance(module, (BaseTunerLayer, ModulesToSaveWrapper)):
            futile = False
            module.enable_adapters(enabled=enabled)

    if futile:
        raise RuntimeError("PEFT model incorrectly initialized")



from pathlib import Path
from typing import Any

import pytest
import torch

from bergson import Attributor, FaissConfig, GradientProcessor, collect_gradients
from bergson.config import IndexConfig


@pytest.mark.skipif(not torch.cuda.is_available(), reason="CUDA not available")
def test_attributor(tmp_path: Path, model, dataset):
    cfg = IndexConfig(run_path=str(tmp_path), token_batch_size=1024)
    cfg.skip_preconditioners = True

    collect_gradients(
        model=model,
        data=dataset,
        processor=GradientProcessor(),
        cfg=cfg,
    )

    attr = Attributor(cfg.partial_run_path, device="cpu", unit_norm=True)

    x = torch.tensor(dataset[0]["input_ids"]).unsqueeze(0)

    with attr.trace(model.base_model, 5) as result:
        model(x, labels=x).loss.backward()
        model.zero_grad()

    assert result.scores[0, 0].item() > 0.99  # Same item
    assert result.scores[0, 1].item() < 0.50  # Different item


@pytest.mark.skipif(not torch.cuda.is_available(), reason="CUDA not available")
def test_faiss(tmp_path: Path, model, dataset):
    pytest.importorskip("faiss")
    dtype: Any = model.dtype
    model.to("cuda")
    dtype = torch.float32 if model.dtype == torch.float32 else torch.float16

    cfg = IndexConfig(run_path=str(tmp_path), token_batch_size=1024)

    collect_gradients(
        model=model,
        data=dataset,
        processor=GradientProcessor(projection_dim=16),
        cfg=cfg,
    )

    attr = Attributor(
        cfg.partial_run_path,
        device="cuda",
        unit_norm=True,
        faiss_cfg=FaissConfig(),
        dtype=dtype,
    )

    x = torch.tensor(dataset[0]["input_ids"]).unsqueeze(0).cuda()

    with attr.trace(model.base_model, 2) as result:
        model(x, labels=x).loss.backward()

        model.zero_grad()

    assert result.scores[0, 0].item() > 0.99  # Same item
    assert result.scores[0, 1].item() < 0.50  # Different item


@pytest.mark.skipif(not torch.cuda.is_available(), reason="CUDA not available")
def test_attributor_reverse(tmp_path: Path, model, dataset):
    """Test that reverse mode returns lowest influence examples."""
    cfg = IndexConfig(run_path=str(tmp_path), token_batch_size=1024)
    cfg.skip_preconditioners = True

    collect_gradients(
        model=model,
        data=dataset,
        processor=GradientProcessor(),
        cfg=cfg,
    )

    attr = Attributor(cfg.partial_run_path, device="cpu", unit_norm=True)

    x = torch.tensor(dataset[0]["input_ids"]).unsqueeze(0)

    # Get normal results (highest influence)
    with attr.trace(model.base_model, 5) as normal_result:
        model(x, labels=x).loss.backward()
        model.zero_grad()

    # Get reverse results (lowest influence)
    with attr.trace(model.base_model, 5, reverse=True) as reverse_result:
        model(x, labels=x).loss.backward()
        model.zero_grad()

    # In reverse mode, the first result should have lower score than normal mode
    assert reverse_result.scores[0, 0].item() < normal_result.scores[0, 0].item()

    # The same item (index 0) should be first in normal mode but not in reverse
    assert normal_result.indices[0, 0].item() == 0  # Same item is top match
    assert reverse_result.indices[0, 0].item() != 0  # Same item is NOT lowest match


@pytest.mark.skipif(not torch.cuda.is_available(), reason="CUDA not available")
def test_faiss_reverse(tmp_path: Path, model, dataset):
    """Test that reverse mode works with FAISS index."""
    pytest.importorskip("faiss")
    dtype: Any = model.dtype
    model.to("cuda")
    dtype = torch.float32 if model.dtype == torch.float32 else torch.float16

    cfg = IndexConfig(run_path=str(tmp_path), token_batch_size=1024)

    collect_gradients(
        model=model,
        data=dataset,
        processor=GradientProcessor(projection_dim=16),
        cfg=cfg,
    )

    attr = Attributor(
        cfg.partial_run_path,
        device="cuda",
        unit_norm=True,
        faiss_cfg=FaissConfig(),
        dtype=dtype,
    )

    x = torch.tensor(dataset[0]["input_ids"]).unsqueeze(0).cuda()

    # Get normal results (highest influence)
    with attr.trace(model.base_model, 2) as normal_result:
        model(x, labels=x).loss.backward()
        model.zero_grad()

    # Get reverse results (lowest influence)
    with attr.trace(model.base_model, 2, reverse=True) as reverse_result:
        model(x, labels=x).loss.backward()
        model.zero_grad()

    # In reverse mode, the first result should have lower score than normal mode
    assert reverse_result.scores[0, 0].item() < normal_result.scores[0, 0].item()

    # The same item (index 0) should be first in normal mode but not in reverse
    assert normal_result.indices[0, 0].item() == 0  # Same item is top match
    assert reverse_result.indices[0, 0].item() != 0  # Same item is NOT lowest match



from contextlib import contextmanager
from pathlib import Path
from typing import Generator

import torch
from torch import Tensor, nn

from bergson.collector.gradient_collectors import TraceCollector
from bergson.data import load_gradients
from bergson.gradients import GradientProcessor
from bergson.query.faiss_index import FaissConfig, FaissIndex
from bergson.utils.utils import numpy_to_tensor


class TraceResult:
    """Result of a .trace() call."""

    def __init__(self):
        # Should be set by the Attributor after a search
        self._indices: Tensor | None = None
        self._scores: Tensor | None = None

    @property
    def indices(self) -> Tensor:
        """The indices of the top-k examples."""
        if self._indices is None:
            raise ValueError("No indices available. Exit the context manager first.")

        return self._indices

    @property
    def scores(self) -> Tensor:
        """The attribution scores of the top-k examples."""
        if self._scores is None:
            raise ValueError("No scores available. Exit the context manager first.")

        return self._scores


class Attributor:
    def __init__(
        self,
        index_path: str | Path,
        device: str = "cpu",
        dtype: torch.dtype = torch.float32,
        unit_norm: bool = False,
        faiss_cfg: FaissConfig | None = None,
    ):
        self.device = device
        self.dtype = dtype
        self.unit_norm = unit_norm
        self.faiss_index = None
        index_path = Path(index_path)

        # Load the gradient processor
        self.processor = GradientProcessor.load(index_path, map_location=device)

        # Load the gradients into a FAISS index
        if faiss_cfg:
            faiss_index_name = (
                f"faiss_{faiss_cfg.index_factory.replace(',', '_')}"
                f"{'_cosine' if unit_norm else ''}"
            )
            faiss_path = index_path / faiss_index_name

            if not (faiss_path / "config.json").exists():
                FaissIndex.create_index(
                    index_path, faiss_path, faiss_cfg, device, unit_norm
                )

            self.faiss_index = FaissIndex(
                faiss_path, device, mmap_index=faiss_cfg.mmap_index
            )
            self.N = self.faiss_index.ntotal
            self.ordered_modules = self.faiss_index.ordered_modules
            return

        # Load the gradients into memory
        mmap = load_gradients(index_path)
        assert mmap.dtype.names is not None
        # Copy gradients into device memory
        self.grads = {
            name: numpy_to_tensor(mmap[name]).to(device=device, dtype=dtype)
            for name in mmap.dtype.names
        }
        self.N = mmap[mmap.dtype.names[0]].shape[0]

        self.ordered_modules = mmap.dtype.names

        if unit_norm:
            norm = torch.cat(
                [self.grads[name] for name in self.ordered_modules], dim=1
            ).norm(dim=1, keepdim=True)

            for name in self.grads:
                # Divide by norm (may create NaN/inf if norm is zero)
                normalized = self.grads[name] / norm
                # Convert NaN/inf to 0 and warn if any were found
                if not torch.isfinite(normalized).all():
                    print(
                        f"Warning: NaN/inf values detected after normalization in "
                        f"{name}, converting to 0"
                    )
                self.grads[name] = torch.nan_to_num(
                    normalized, nan=0.0, posinf=0.0, neginf=0.0
                )

    def search(
        self,
        queries: dict[str, Tensor],
        k: int | None,
        modules: set[str] | None = None,
        reverse: bool = False,
    ):
        """
        Search for the `k` nearest examples in the index based on the query or queries.

        Args:
            queries: The query tensor of shape [..., d].
            k: The number of nearest examples to return for each query.
            module: The name of the module to search for. If `None`,
                all modules will be searched.
            reverse: If True, return the lowest influence examples instead of highest.

        Returns:
            A namedtuple containing the top `k` indices and inner products for each
            query. Both have shape [..., k].
        """
        q = {
            name: queries[name].to(self.device, self.dtype)
            for name in self.ordered_modules
        }

        if self.unit_norm:
            norm = torch.cat(list(q.values()), dim=1).norm(dim=1, keepdim=True)

            for name in q:
                q[name] /= norm + torch.finfo(norm.dtype).eps

        if self.faiss_index:
            if modules:
                raise NotImplementedError(
                    "FAISS index does not implement module-specific search."
                )

            q = (
                torch.cat([q[name] for name in self.ordered_modules], dim=1)
                .cpu()
                .numpy()
            )

            distances, indices = self.faiss_index.search(q, k, reverse=reverse)

            return torch.from_numpy(distances), torch.from_numpy(indices)

        if modules:
            modules = set([name for name in self.ordered_modules if name in modules])
        else:
            modules = set(self.ordered_modules)

        k = min(k or self.N, self.N)

        scores = torch.stack(
            [q[name] @ self.grads[name].mT for name in modules], dim=-1
        ).sum(-1)

        return torch.topk(scores, k, largest=not reverse)  # type: ignore

    @contextmanager
    def trace(
        self,
        module: nn.Module,
        k: int | None,
        *,
        precondition: bool = False,
        modules: set[str] | None = None,
        reverse: bool = False,
    ) -> Generator[TraceResult, None, None]:
        """
        Context manager to trace the gradients of a module and return the
        corresponding Attributor instance.

        Args:
            module: The module to trace.
            k: The number of nearest examples to return.
            precondition: Whether to apply preconditioning.
            modules: The modules to trace. If None, all modules will be traced.
            reverse: If True, return the lowest influence examples instead of highest.
        """

        result = TraceResult()

        collector = TraceCollector(
            model=module,
            processor=self.processor,
            precondition=precondition,
            target_modules=modules,
            device=self.device,
            dtype=self.dtype,
        )
        mod_grads = collector.mod_grads
        with collector:
            yield result

        if not mod_grads:
            raise ValueError("No grads collected. Did you forget to call backward?")

        queries = {
            name: torch.cat(mod_grads[name], dim=1)
            for name in self.ordered_modules
            if name in mod_grads
        }

        if any(q.isnan().any() for q in queries.values()):
            raise ValueError("NaN found in queries.")

        result._scores, result._indices = self.search(queries, k, modules, reverse)



import math
import random
from dataclasses import dataclass, field

import torch
import torch.distributed as dist
import torch.nn as nn
from datasets import Dataset
from jaxtyping import Float
from torch import Tensor
from transformers import PreTrainedModel

from bergson.collector.collector import CollectorComputer, HookCollectorBase
from bergson.config import IndexConfig
from bergson.gradients import (
    AdafactorNormalizer,
    AdamNormalizer,
    LayerAdapter,
    Normalizer,
)
from bergson.process_preconditioners import process_preconditioners
from bergson.utils.utils import assert_type, get_gradient_dtype


@dataclass(kw_only=True)
class NormalizerCollector(HookCollectorBase):
    """
    Collects per-sample gradients from model layers and writes them to disk.

    - For each forward/backward hook, we compute the the gradient or a low-rank
    approximation via random projections, if cfg.projection_dim is set.
    - Supports also normalization via Adam or Adafactor normalizers.
    - Uses Builder for index construction and gradient saving.
    - Also supports Scorer for on-the-fly scoring of gradients.
    """

    data: Dataset
    """The dataset being processed."""

    cfg: IndexConfig
    """Configuration for gradient index."""

    normalizers: dict[str, Normalizer] = field(default_factory=dict)

    mod_grads: dict = field(default_factory=dict)
    """Temporary storage for gradients during a batch, keyed by module name."""

    def __init__(self, *args, **kwargs):
        self.data = assert_type(Dataset, kwargs["data"])
        self.cfg = assert_type(IndexConfig, kwargs["cfg"])
        self.normalizers = {}
        self.mod_grads = {}

        self.callback = (
            self.adafactor_update
            if self.cfg.normalizer == "adafactor"
            else self.adam_update
        )

        # Extract parent class arguments
        parent_kwargs = {
            k: v
            for k, v in kwargs.items()
            if k
            in {
                "model",
                "filter_modules",
                "target_modules",
                "processor",
                "attention_cfgs",
            }
        }
        parent_kwargs["filter_modules"] = self.cfg.filter_modules

        super().__init__(*args, **parent_kwargs)

    def adafactor_update(self, name: str, g: torch.Tensor):
        # We follow the tensor2tensor implementation of Adafactor, which
        # takes the mean rather than summing over the rows and columns.
        # row: mean over columns, shape [O]
        sq = g.float().square_().sum(0)
        row_acc = sq.mean(dim=1)
        # col: mean over rows,    shape [I]
        col_acc = sq.mean(dim=0)

        if (normalizer := self.normalizers.get(name)) is None:
            # initialize accumulators at zero
            self.normalizers[name] = normalizer = AdafactorNormalizer(
                torch.zeros_like(row_acc),
                torch.zeros_like(col_acc),
            )
        else:
            assert isinstance(normalizer, AdafactorNormalizer)

        # in‐place accumulate
        normalizer.row.add_(row_acc)
        normalizer.col.add_(col_acc)

    def adam_update(self, name: str, g: torch.Tensor):
        sq = g.square_().float().sum(0)

        # initialize accumulators at zero
        if (normalizer := self.normalizers.get(name)) is None:
            self.normalizers[name] = normalizer = AdamNormalizer(torch.zeros_like(sq))
        else:
            assert isinstance(normalizer, AdamNormalizer)

        # in‐place accumulate
        normalizer.avg_sq.add_(sq)

    def setup(self) -> None:
        """
        Initialize collector state.

        Sets up a Builder for gradient storage if not using a Scorer.
        """
        assert isinstance(
            self.model.device, torch.device
        ), "Model device is not set correctly"
        if self.cfg.include_bias and self.processor.normalizers is not None:
            raise NotImplementedError(
                "Bias with normalizers not supported yet, "
                "consider disabling bias inclusion for now."
            )

        self.save_dtype = get_gradient_dtype(self.model)
        self.lo = torch.finfo(self.save_dtype).min
        self.hi = torch.finfo(self.save_dtype).max

    def forward_hook(self, module: nn.Module, a: Float[Tensor, "N S I"]) -> None:
        """
        Cache activations for gradient computation with normalizer preprocessing
        and compress via random projection if configured.
        Stores result in module._inputs for use in backward_hook.
        """
        p = self.processor.projection_dim
        name = assert_type(str, module._name)
        i = getattr(module, LayerAdapter.in_attr(module))
        normalizer = self.processor.normalizers.get(name)

        if isinstance(normalizer, AdamNormalizer):
            module._inputs = a
            return
        if isinstance(normalizer, AdafactorNormalizer):
            a_factor = normalizer.col.add(1e-30)
            a_factor = a_factor.rsqrt()
            a = a * a_factor.type_as(a)  # [N, S, I] * [I] → [N, S, I]

        if module._has_bias:
            # Append ones to activation for bias term
            ones = torch.ones(a.size(0), a.size(1), 1, device=a.device, dtype=a.dtype)
            a = torch.cat([a, ones], dim=-1)
            i = i + 1
            setattr(module, LayerAdapter.in_attr(module), i)
        if p is not None:
            a_projection = self.projection(name, p, i, "right", a.device, a.dtype).T
            a = a @ a_projection  # type: ignore
        # set module._inputs to a
        module._inputs = a

    @HookCollectorBase.split_attention_heads
    def backward_hook(self, module: nn.Module, g: Float[Tensor, "N S O"]):
        """
        Compute per-sample gradient and store in mod_grads.

        Computes gradient as outer product g.T @ a (again with optional projection and
        normalization).
        """
        a = module._inputs  # [N, S, I/q]

        assert isinstance(a, torch.Tensor), "Activation cache missing for module"
        name = assert_type(str, module._name)
        p = self.processor.projection_dim
        i = getattr(module, LayerAdapter.in_attr(module))
        o = getattr(module, LayerAdapter.out_attr(module))
        normalizer = self.processor.normalizers.get(name)

        if isinstance(normalizer, AdamNormalizer):
            full_gradient = g.mT @ a  # [N, O, S] @ [N, S, I] → [N, O, I]
            P = normalizer.normalize_(full_gradient)
            if p is not None:
                g_projection = self.projection(name, p, o, "left", g.device, g.dtype)
                a_projection = self.projection(name, p, i, "right", g.device, g.dtype).T
                P = g_projection @ P @ a_projection
        else:
            if isinstance(normalizer, AdafactorNormalizer):
                g_factor = normalizer.row.add(1e-30)
                g_factor = g_factor.mean().sqrt() * g_factor.rsqrt()
                g = g * g_factor.type_as(g)  # [N, S, O] * [O] → [N, S, O]

            if p is not None:
                g_projection = self.projection(name, p, o, "left", g.device, g.dtype)
                g = g @ g_projection.T  # [N, S, p]

            P = g.mT @ a  # [N, O/p, S] @ [N, S, I/q] → [N, O/p, I/q]

        P = P.flatten(1).clamp_(self.lo, self.hi)

        if not self.cfg.skip_preconditioners:
            P = P.float()
            if name in self.processor.preconditioners:
                self.processor.preconditioners[name].addmm_(P.mT, P)
            else:
                self.processor.preconditioners[name] = P.mT @ P

        # self.mod_grads[name] = P.to(dtype=self.save_dtype)
        self.callback(name, P)

        del module._inputs

    def process_batch(self, indices: list[int], **kwargs):
        """Process collected gradients for a batch."""
        self.mod_grads.clear()

    def teardown(self):
        """
        Finalize normalizer collection.
        """
        grad_sizes = {name: math.prod(s) for name, s in self.shapes().items()}
        if self.processor.preconditioners:
            process_preconditioners(
                self.processor,
                self.processor.preconditioners,
                len(self.data),
                grad_sizes,
                self.rank,
            )

        if self.rank == 0:
            self.processor.save(self.cfg.partial_run_path)


def fit_normalizers(
    model: PreTrainedModel,
    data: Dataset,
    cfg: IndexConfig,
    batches: list[list[int]],
    *,
    target_modules: set[str] | None = None,
) -> dict[str, Normalizer]:
    """
    Estimate the second moments of the model's gradients using a subset of the dataset.
    """
    # Just to make the pbar more accurate
    rng = random.Random(0)
    rng.shuffle(batches)

    collector = NormalizerCollector(
        model=model.base_model,  # type: ignore
        data=data,
        cfg=cfg,
        target_modules=target_modules,
    )
    computer = CollectorComputer(
        model=model,
        data=data,
        collector=collector,
        cfg=cfg,
    )
    computer.run_with_collector_hooks(desc="Estimating normalizers")

    normalizers = collector.normalizers

    # Divide by the number of documents processed and average across all ranks
    for normalizer in normalizers.values():
        if isinstance(normalizer, AdamNormalizer):
            normalizer.avg_sq.div_(len(data))

            if dist.is_initialized():
                dist.all_reduce(normalizer.avg_sq, op=dist.ReduceOp.AVG)

        elif isinstance(normalizer, AdafactorNormalizer):
            normalizer.row.div_(len(data))
            normalizer.col.div_(len(data))

            if dist.is_initialized():
                dist.all_reduce(normalizer.row, op=dist.ReduceOp.AVG)
                dist.all_reduce(normalizer.col, op=dist.ReduceOp.AVG)

    return normalizers



# test_finite_diff.py
import pytest
import torch
import torch.nn as nn

from bergson import FiniteDiff


def make_model():
    lin1 = nn.Linear(5, 3, bias=False)
    lin2 = nn.Linear(3, 4, bias=False)
    return nn.Sequential(lin1, lin2)


def clone_params(model):
    "Return a dict of *detached* clones of all parameters."
    return {n: p.detach().clone() for n, p in model.named_parameters()}


def forward_loss(model, x):
    """Dummy forward that attaches a .loss attribute for FiniteDiff."""
    y = model(x)
    return (y**2).mean()


@pytest.mark.parametrize("step_size", [0.05, -0.05])  # check both signs
def test_reversibility(step_size):
    torch.manual_seed(0)

    model = make_model()
    fd = FiniteDiff(model)

    x = torch.randn(7, 5)
    loss = forward_loss(model, x)

    # 1. backward & remember grads/params
    model.zero_grad(set_to_none=True)
    loss.backward()

    grads = {n: p.grad.detach().clone() for n, p in model.named_parameters()}  # type: ignore
    orig = clone_params(model)

    # 2. compute finite differences
    fd.store(step_size)

    # after compute: parameters should be unchanged, grads gone
    for n, p in model.named_parameters():
        assert torch.allclose(p, orig[n]), f"{n} changed during compute"
        assert p.grad is None, f"{n}.grad not cleared"

    # 3. byte-level swap in context manager
    with fd.apply():
        for n, p in model.named_parameters():
            if n in fd.params:
                expected = orig[n] + step_size * grads[n]
                assert torch.allclose(p, expected), f"{n} not updated inside ctx"

    # 4. on exit, parameters must be *exactly* as before (bit-wise)
    for n, p in model.named_parameters():
        assert torch.equal(p, orig[n]), f"{n} not restored exactly"

    # 5. cleanup code-path
    fd.clear()
    assert fd.params == {}, "params dict not cleared"


def test_multiple_swaps():
    """Calling swap() twice manually should be a no-op overall."""
    model = make_model()
    fd = FiniteDiff(model)

    # Create grads
    x = torch.randn(3, 5)
    out = forward_loss(model, x)
    model.zero_grad()
    out.backward()

    fd.store(0.1)
    before = clone_params(model)

    fd.swap()  # first swap --> weights changed
    fd.swap()  # second swap --> back to original

    after = clone_params(model)
    for n in before:
        assert torch.equal(before[n], after[n]), f"{n} not identical after 2 swaps"



import math

import torch
from torch import Tensor


def optimal_linear_shrinkage(S_n: Tensor, n: int | Tensor) -> Tensor:
    """Optimal linear shrinkage for a sample covariance matrix or batch thereof.

    Given a sample covariance matrix `S_n` of shape (*, p, p) and a sample size `n`,
    this function computes the optimal shrinkage coefficients `alpha` and `beta`, then
    returns the covariance estimate `alpha * S_n + beta * Sigma0`, where ``Sigma0` is
    an isotropic covariance matrix with the same trace as `S_n`.

    The formula is distribution-free and asymptotically optimal in the Frobenius norm
    among all linear shrinkage estimators as the dimensionality `p` and sample size `n`
    jointly tend to infinity, with the ratio `p / n` converging to a finite positive
    constant `c`. The derivation is based on Random Matrix Theory and assumes that the
    underlying distribution has finite moments up to 4 + eps, for some eps > 0.

    See "On the Strong Convergence of the Optimal Linear Shrinkage Estimator for Large
    Dimensional Covariance Matrix" <https://arxiv.org/abs/1308.2608> for details.

    Args:
        S_n: Sample covariance matrices of shape (*, p, p).
        n: Sample size.
    """
    p = S_n.shape[-1]
    assert n > 1 and S_n.shape[-2:] == (p, p)

    # TODO: Make this configurable, try using diag(S_n) or something
    eye = torch.eye(p, dtype=S_n.dtype, device=S_n.device).expand_as(S_n)
    trace_S = trace(S_n)
    sigma0 = eye * trace_S / p

    sigma0_norm_sq = sigma0.pow(2).sum(dim=(-2, -1), keepdim=True)
    S_norm_sq = S_n.pow(2).sum(dim=(-2, -1), keepdim=True)

    prod_trace = trace(S_n @ sigma0)
    top = trace_S.pow(2) * sigma0_norm_sq / n
    bottom = S_norm_sq * sigma0_norm_sq - prod_trace**2

    # Epsilon prevents dividing by zero for the zero matrix. In that case we end up
    # setting alpha = 0, beta = 1, but it doesn't matter since we're shrinking toward
    # tr(0)*I = 0, so it's a no-op.
    eps = torch.finfo(S_n.dtype).eps

    # Ensure that alpha and beta are in [0, 1] and thereby ensure that the resulting
    # covariance matrix is positive semi-definite.
    alpha = torch.clamp(1 - (top + eps) / (bottom + eps), min=0, max=1)
    beta = (1 - alpha) * (prod_trace + eps) / (sigma0_norm_sq + eps)

    return alpha * S_n + beta * sigma0


@torch.compile
def psd_rsqrt(A: Tensor) -> Tensor:
    """Efficiently compute the p.s.d. pseudoinverse sqrt of p.s.d. matrix `A`."""
    L, U = torch.linalg.eigh(A)
    L = L[..., None, :].clamp_min(0.0)

    # We actually compute the pseudo-inverse here for numerical stability.
    # Use the same heuristic as `torch.linalg.pinv` to determine the tolerance.
    thresh = L[..., None, -1] * A.shape[-1] * torch.finfo(A.dtype).eps
    rsqrt = U * torch.where(L > thresh, L.rsqrt(), 0.0) @ U.mH

    return rsqrt


def compute_damped_inverse(
    H: Tensor,
    damping_factor: float = 0.1,
    dtype: torch.dtype = torch.float64,
    regularizer: Tensor | None = None,
) -> Tensor:
    """Compute H^(-1) with damping for numerical stability.

    Uses eigendecomposition to compute the inverse of a positive semi-definite
    matrix with adaptive damping based on the matrix's mean absolute value.

    Args:
        H: Positive semi-definite matrix to invert.
        damping_factor: Multiplier for the damping term (default: 0.1).
        dtype: Dtype for intermediate computation (default: float64 for stability).
        regularizer: Optional matrix to use as regularizer instead of identity.
            If provided, computes inv(H + damping_factor * regularizer).
            If None (default), uses scaled identity:
            inv(H + damping_factor * |H|_mean * I).

    Returns:
        The damped inverse H^(-1) in the original dtype of H.
    """
    original_dtype = H.dtype
    H = H.to(dtype=dtype)
    if regularizer is not None:
        regularizer = regularizer.to(dtype=dtype, device=H.device)
        H = H + damping_factor * regularizer
    else:
        damping_val = damping_factor * H.abs().mean()
        H = H + damping_val * torch.eye(H.shape[0], device=H.device, dtype=H.dtype)
    eigval, eigvec = torch.linalg.eigh(H)
    return (eigvec * (1.0 / eigval) @ eigvec.mT).to(original_dtype)


def trace(matrices: Tensor) -> Tensor:
    """Version of `torch.trace` that works for batches of matrices."""
    diag = torch.linalg.diagonal(matrices)
    return diag.sum(dim=-1, keepdim=True).unsqueeze(-1)


def reshape_to_nearest_square(a: torch.Tensor) -> torch.Tensor:
    """
    Reshape a 2-D (or any-D) tensor into the *most square* 2-D shape
    that preserves the total number of elements.

    Returns
    -------
    out   : reshaped tensor (view when possible)
    shape : tuple (rows, cols) that was chosen
    """
    n = math.prod(a.shape[-2:])
    if n == 0:
        raise ValueError("empty tensor")

    # search divisors closest to sqrt(n)
    root = math.isqrt(n)
    cols, rows = None, None
    for d in range(root, 0, -1):
        if n % d == 0:
            rows = d
            cols = n // d
            break

    if rows is None or cols is None:
        raise ValueError("could not find a valid shape for the tensor")

    return a.reshape(*a.shape[:-2], rows, cols)



import json
from pathlib import Path
from time import perf_counter
from types import ModuleType
from typing import TYPE_CHECKING, Protocol

import numpy as np
import psutil
import torch
from numpy.lib.recfunctions import structured_to_unstructured
from numpy.typing import NDArray
from tqdm import tqdm

from bergson.config import FaissConfig

if TYPE_CHECKING:
    import faiss  # noqa: F401


class Index(Protocol):
    """Protocol for searchable FAISS index."""

    def search(self, x: NDArray, k: int) -> tuple[NDArray, NDArray]: ...
    @property
    def ntotal(self) -> int: ...
    def train(self, x: NDArray) -> None: ...
    def add(self, x: NDArray) -> None: ...


def validate_ram(batch: NDArray, max_shard_size: int):
    """
    Estimate the peak RAM footprint for the largest shard and assert it fits.

    Parameters
    ----------
    batch : NDArray
        Sample of gradients used to approximate bytes per gradient vector.
    max_shard_size : int
        Maximum number of gradients that will be grouped into any shard.
    """
    available_ram_gb = psutil.virtual_memory().available / (1024**3)

    bytes_per_grad = batch.nbytes / batch.shape[0]
    estimated_shard_ram_gb = (max_shard_size * bytes_per_grad) / (1024**3)

    print(f"Estimated RAM required for largest shard: {estimated_shard_ram_gb} GB")
    print(f"Available RAM: {available_ram_gb} GB")

    assert estimated_shard_ram_gb <= available_ram_gb, (
        "Not enough RAM to build index."
        "Increase the number of shards to reduce peak RAM usage."
    )


def normalize_grads(
    grads: NDArray,
    device: str,
    batch_size: int,
) -> NDArray:
    """
    Normalize gradients to unit norm in batches to keep GPU memory bounded.

    Parameters
    ----------
    grads : NDArray
        Array containing all gradient vectors to normalize.
    device : str
        Device identifier understood by PyTorch (e.g., ``\"cuda:0\"`` or ``\"cpu\"``).
    batch_size : int
        Number of gradients processed per batch before writing back to host memory.
    """
    normalized_grads = np.zeros_like(grads).astype(grads.dtype)

    for i in range(0, grads.shape[0], batch_size):
        batch = torch.from_numpy(grads[i : i + batch_size]).to(device)
        normalized_grads[i : i + batch_size] = (
            (batch / batch.norm(dim=1, keepdim=True)).cpu().numpy()
        )

    return normalized_grads


def gradients_loader(root_dir: Path):
    """
    Yield memory-mapped gradient shards stored under ``root_dir``.

    Handles both single-shard exports (``info.json`` directly under ``root_dir``)
    and multi-shard layouts (directories named ``*shard*``). Each yielded object
    stays memory-mapped so downstream code can stream slices without excessive RAM.
    """

    def load_shard(shard_dir: Path) -> np.memmap:
        with (shard_dir / "info.json").open("r") as f:
            info = json.load(f)

        return np.memmap(
            shard_dir / "gradients.bin",
            dtype=info["dtype"],
            mode="r",
            shape=(info["num_grads"],),
        )

    if (root_dir / "info.json").exists():
        yield load_shard(root_dir)
    else:
        for path in sorted(root_dir.iterdir()):
            if "shard" in path.name:
                yield load_shard(path)


def _require_faiss() -> ModuleType:
    """Import faiss at runtime and raise an error if missing."""

    try:
        import faiss as faiss_module  # type: ignore[import]
    except ImportError as e:
        raise ImportError("Faiss not found, run `pip install faiss-gpu-cu12`") from e

    return faiss_module


def index_to_device(index: Index, device: str) -> Index:
    """
    Move a FAISS index between CPU and GPU devices, optionally sharding.

    Parameters
    ----------
    index : Index
        Existing FAISS index instance.
    device : str
        Destination device string, e.g. ``\"cpu\"``, ``\"cuda\"``, or ``\"cuda:1\"``.
    """
    faiss = _require_faiss()

    if device != "cpu":
        gpus = (
            list(range(torch.cuda.device_count()))
            if device == "cuda"
            else [int(device.split(":")[1])]
        )

        try:
            options = faiss.GpuMultipleClonerOptions()
        except AttributeError as e:
            raise ImportError(
                "Faiss not found, you may have faiss-cpu installed instead "
                "of faiss-gpu with `pip install faiss-gpu-cu12`..."
            ) from e

        options.shard = True
        return faiss.index_cpu_to_gpus_list(index, options, gpus=gpus)

    return faiss.index_gpu_to_cpu(index)


class FaissIndex:
    """Sharded FAISS index for efficient nearest neighbor search."""

    shards: list[Index]

    faiss_cfg: FaissConfig

    ordered_modules: list[str]

    def __init__(self, path: Path, device: str, mmap_index: bool):
        faiss = _require_faiss()

        config_path = Path(path) / "config.json"

        if not config_path.exists():
            raise FileNotFoundError(
                f"FAISS index configuration not found at {config_path}."
                "Run `FaissIndex.create_index` to create the index."
            )

        with open(config_path) as f:
            config = json.load(f)

        self.unit_norm = config["unit_norm"]
        self.ordered_modules = config["ordered_modules"]
        self.faiss_cfg = FaissConfig(**config["faiss_cfg"])

        shard_paths = sorted(
            (c for c in path.glob("*.faiss") if c.stem.isdigit()),
            key=lambda p: int(p.stem),
        )

        shards = []
        for shard_path in shard_paths:
            shard = faiss.read_index(
                str(shard_path),
                faiss.IO_FLAG_MMAP | faiss.IO_FLAG_READ_ONLY,
            )
            if not mmap_index:
                shard = index_to_device(shard, device)

            shards.append(shard)

        self.shards = shards

    @staticmethod
    def create_index(
        gradients_path: Path,
        faiss_path: Path,
        faiss_cfg: FaissConfig,
        device: str,
        unit_norm: bool,
    ):
        faiss = _require_faiss()

        print("Building FAISS index...")
        start = perf_counter()

        faiss_path.mkdir(exist_ok=True, parents=True)

        # Write the gradients into an on-disk FAISS index
        if (gradients_path / "info.json").exists():
            info_paths = [gradients_path / "info.json"]
        else:
            info_paths = [
                shard_path / "info.json"
                for shard_path in gradients_path.iterdir()
                if (shard_path / "info.json").exists()
            ]

        assert info_paths, f"No gradient metadata found under {gradients_path}"

        total_grads = sum(
            [json.load(open(info_path))["num_grads"] for info_path in info_paths]
        )

        assert faiss_cfg.num_shards <= total_grads and faiss_cfg.num_shards > 0

        # Set the number of grads for each faiss index shard
        base_shard_size = total_grads // faiss_cfg.num_shards
        remainder = total_grads % faiss_cfg.num_shards
        shard_sizes = [base_shard_size] * (faiss_cfg.num_shards)
        shard_sizes[-1] += remainder

        # Verify all gradients will be consumed
        assert (
            sum(shard_sizes) == total_grads
        ), f"Shard sizes {shard_sizes} don't sum to total_grads {total_grads}"

        dl = gradients_loader(gradients_path)
        buffer: list[NDArray] = []
        buffer_size = 0
        shard_idx = 0

        def build_shard_from_buffer(
            buffer_parts: list[NDArray], shard_idx: int
        ) -> None:
            shard_path = faiss_path / f"{shard_idx}.faiss"
            if shard_path.exists():
                print(f"Shard {shard_idx} already exists, skipping...")
                return
            else:
                print(f"Building shard {shard_idx}...")

            grads_chunk = np.concatenate(buffer_parts, axis=0)
            buffer_parts.clear()

            index = faiss.index_factory(
                grads_chunk.shape[1],
                faiss_cfg.index_factory,
                faiss.METRIC_INNER_PRODUCT,
            )
            index = index_to_device(index, device)
            if faiss_cfg.max_train_examples is not None:
                train_examples = min(faiss_cfg.max_train_examples, grads_chunk.shape[0])
            else:
                train_examples = grads_chunk.shape[0]
            index.train(grads_chunk[:train_examples])
            index.add(grads_chunk)

            del grads_chunk

            index = index_to_device(index, "cpu")
            faiss.write_index(index, str(shard_path))

        ordered_modules = []
        for i, grads in enumerate(tqdm(dl, desc="Loading gradients")):
            if i == 0:
                ordered_modules = list(grads.dtype.names or [])

            grads = structured_to_unstructured(grads)

            if i == 0:
                validate_ram(grads, shard_sizes[-1])

            if unit_norm:
                grads = normalize_grads(grads, device, faiss_cfg.batch_size)

            batch_idx = 0
            batch_size = grads.shape[0]
            while batch_idx < batch_size and shard_idx < faiss_cfg.num_shards:
                remaining_in_shard = shard_sizes[shard_idx] - buffer_size
                take = min(remaining_in_shard, batch_size - batch_idx)

                if take > 0:
                    buffer.append(grads[batch_idx : batch_idx + take])
                    buffer_size += take
                    batch_idx += take

                if buffer_size == shard_sizes[shard_idx]:
                    build_shard_from_buffer(buffer, shard_idx)
                    buffer = []
                    buffer_size = 0
                    shard_idx += 1

            del grads

        # Write the configuration to disk
        with open(faiss_path / "config.json", "w") as f:
            json.dump(
                {
                    "faiss_cfg": faiss_cfg.__dict__,
                    "gradients_path": str(gradients_path),
                    "device": device,
                    "unit_norm": unit_norm,
                    "ordered_modules": ordered_modules,
                },
                f,
                indent=2,
            )

        print(f"Built index in {(perf_counter() - start) / 60:.2f} minutes.")

    def search(
        self, q: NDArray, k: int | None, reverse: bool = False
    ) -> tuple[NDArray, NDArray]:
        """
        Perform a nearest neighbor search on the index.

        If fewer than `k` items are found, invalid items will be returned
        with index -1 and a maximum-valued negative distance. If `k` is
        `None`, all available items are returned.

        Args:
            q: Query vectors of shape [num_queries, dim].
            k: Number of results to return per query.
            reverse: If True, return lowest influence examples instead of highest.
        """
        shard_distances = []
        shard_indices = []
        offset = 0

        # For reverse mode with FAISS, we need to fetch all results and then
        # select the lowest scores
        fetch_k = self.ntotal if reverse else k

        for shard in self.shards:
            if hasattr(shard, "nprobe"):
                shard.nprobe = self.faiss_cfg.nprobe  # type: ignore

            distances, indices = shard.search(q, fetch_k or shard.ntotal)

            indices += offset
            offset += shard.ntotal

            shard_distances.append(distances)
            shard_indices.append(indices)

        distances = np.concatenate(shard_distances, axis=1)
        indices = np.concatenate(shard_indices, axis=1)

        # Rerank results overfetched from multiple shards or for reverse mode
        if len(self.shards) > 1 or reverse:
            if reverse:
                # For reverse mode, sort ascending (lowest first) and take first k
                topk_indices = np.argsort(distances, axis=1)[:, : k or self.ntotal]
            else:
                # For normal mode, sort descending (highest first) and take first k
                topk_indices = np.argsort(-distances, axis=1)[:, : k or self.ntotal]
            indices = indices[np.arange(indices.shape[0])[:, None], topk_indices]
            distances = distances[np.arange(distances.shape[0])[:, None], topk_indices]

        return distances, indices

    @property
    def ntotal(self) -> int:
        return sum(shard.ntotal for shard in self.shards)

    @property
    def nprobe(self) -> int:
        return self.faiss_cfg.nprobe

    @nprobe.setter
    def nprobe(self, value: int) -> None:
        for shard in self.shards:
            if hasattr(shard, "nprobe"):
                shard.nprobe = value  # type: ignore



from __future__ import annotations

import math
from collections import defaultdict
from dataclasses import dataclass
from dataclasses import field as dc_field

import torch
import torch.distributed as dist
from datasets import Dataset
from jaxtyping import Float
from torch import Tensor, nn

from bergson.collector.collector import HookCollectorBase
from bergson.config import IndexConfig
from bergson.gradients import AdafactorNormalizer, AdamNormalizer, LayerAdapter
from bergson.process_preconditioners import process_preconditioners
from bergson.utils.utils import assert_type, get_gradient_dtype


@dataclass
class InMemoryCollector(HookCollectorBase):
    """Simple collector that accumulates gradients in memory for benchmarking."""

    gradients: dict[str, list[torch.Tensor]] = dc_field(
        default_factory=lambda: defaultdict(list)
    )
    """Accumulated gradients keyed by module name."""

    def __init__(self, *args, **kwargs):
        self.data = assert_type(Dataset, kwargs["data"])
        self.cfg = assert_type(IndexConfig, kwargs["cfg"])

        self.reduce_cfg = kwargs.get("reduce_cfg", None)
        self.builder = kwargs.get("builder", None)
        self.scorer = kwargs.get("scorer", None)

        # Extract parent class arguments
        parent_kwargs = {
            k: v
            for k, v in kwargs.items()
            if k
            in {
                "model",
                "filter_modules",
                "target_modules",
                "processor",
                "attention_cfgs",
            }
        }
        parent_kwargs["filter_modules"] = self.cfg.filter_modules

        super().__init__(*args, **parent_kwargs)

    def setup(self) -> None:
        """Initialize gradient storage."""
        assert isinstance(
            self.model.device, torch.device
        ), "Model device is not set correctly"
        if self.cfg.include_bias and self.processor.normalizers is not None:
            raise NotImplementedError(
                "Bias with normalizers not supported yet, "
                "consider disabling bias inclusion for now."
            )

        self.save_dtype = get_gradient_dtype(self.model)
        self.lo = torch.finfo(self.save_dtype).min
        self.hi = torch.finfo(self.save_dtype).max

        self.per_doc_losses = torch.full(
            (len(self.data),),
            device=self.model.device,
            dtype=torch.float32,
            fill_value=0.0,
        )

        self.gradients = defaultdict(list)

    def teardown(self) -> None:
        """No cleanup needed for in-memory storage."""
        assert isinstance(
            self.cfg, IndexConfig
        ), "cfg is required for GradientCollector"  # pleasing type checker
        if dist.is_initialized():
            dist.reduce(self.per_doc_losses, dst=0)

        grad_sizes = {name: math.prod(s) for name, s in self.shapes().items()}
        if self.processor.preconditioners:
            process_preconditioners(
                self.processor,
                self.processor.preconditioners,
                len(self.data),
                grad_sizes,
                self.rank,
            )

    def forward_hook(self, module: nn.Module, a: Float[Tensor, "N S I"]) -> None:
        """Store activations for gradient computation."""
        p = self.processor.projection_dim
        name = assert_type(str, module._name)
        i = getattr(module, LayerAdapter.in_attr(module))
        normalizer = self.processor.normalizers.get(name)

        if isinstance(normalizer, AdamNormalizer):
            module._inputs = a
            return
        if isinstance(normalizer, AdafactorNormalizer):
            a_factor = normalizer.col.add(1e-30)
            a_factor = a_factor.rsqrt()
            a = a * a_factor.type_as(a)  # [N, S, I] * [I] → [N, S, I]

        if module._has_bias:
            # Append ones to activation for bias term
            ones = torch.ones(a.size(0), a.size(1), 1, device=a.device, dtype=a.dtype)
            a = torch.cat([a, ones], dim=-1)
            i = i + 1
            setattr(module, LayerAdapter.in_attr(module), i)
        if p is not None:
            a_projection = self.projection(name, p, i, "right", a.device, a.dtype).T
            a = a @ a_projection  # type: ignore
        # set module._inputs to a
        module._inputs = a

    def backward_hook(self, module: nn.Module, g: Float[Tensor, "N S O"]) -> None:
        """Compute and store per-sample gradient."""
        a = module._inputs  # [N, S, I/q]

        assert isinstance(a, torch.Tensor), "Activation cache missing for module"
        name = assert_type(str, module._name)
        p = self.processor.projection_dim
        i = getattr(module, LayerAdapter.in_attr(module))
        o = getattr(module, LayerAdapter.out_attr(module))
        normalizer = self.processor.normalizers.get(name)

        if isinstance(normalizer, AdamNormalizer):
            full_gradient = g.mT @ a  # [N, O, S] @ [N, S, I] → [N, O, I]
            P = normalizer.normalize_(full_gradient)
            if p is not None:
                g_projection = self.projection(name, p, o, "left", g.device, g.dtype)
                a_projection = self.projection(name, p, i, "right", g.device, g.dtype).T
                P = g_projection @ P @ a_projection
        else:
            if isinstance(normalizer, AdafactorNormalizer):
                g_factor = normalizer.row.add(1e-30)
                g_factor = g_factor.mean().sqrt() * g_factor.rsqrt()
                g = g * g_factor.type_as(g)  # [N, S, O] * [O] → [N, S, O]

            if p is not None:
                g_projection = self.projection(name, p, o, "left", g.device, g.dtype)
                g = g @ g_projection.T  # [N, S, p]

            P = g.mT @ a  # [N, O/p, S] @ [N, S, I/q] → [N, O/p, I/q]

        P = P.flatten(1).clamp_(self.lo, self.hi)

        if not self.cfg.skip_preconditioners:
            P = P.float()
            if name in self.processor.preconditioners:
                self.processor.preconditioners[name].addmm_(P.mT, P)
            else:
                self.processor.preconditioners[name] = P.mT @ P

        # Store on CPU
        self.gradients[name].append(P.cpu())

        del module._inputs  # type: ignore

    def process_batch(self, indices: list[int], **kwargs) -> None:
        losses = kwargs.get("losses")
        assert losses is not None, "losses must be provided in kwargs"

        if self.scorer is not None:
            # Score the most recent gradients in each list
            mod_grads = {name: grads[-1] for name, grads in self.gradients.items()}
            self.scorer(indices, mod_grads)

        self.per_doc_losses[indices] = losses.detach().type_as(self.per_doc_losses)



import functools
import hashlib
import os
from abc import ABC, abstractmethod
from contextlib import ContextDecorator, nullcontext
from dataclasses import astuple, dataclass, field
from fnmatch import fnmatchcase
from typing import Callable, Literal, Mapping, Optional

import numpy as np
import torch
import torch.distributed as dist
import torch.nn as nn
import torch.nn.functional as F
from datasets import Dataset
from jaxtyping import Float
from torch import Tensor
from torch.profiler import (
    ProfilerActivity,
    profile,
    record_function,
    schedule,
    tensorboard_trace_handler,
)
from torch.utils.hooks import RemovableHandle
from tqdm.auto import tqdm
from transformers import PreTrainedModel

from bergson.config import AttentionConfig, HessianConfig, IndexConfig
from bergson.data import pad_and_tensor
from bergson.gradients import (
    GradientProcessor,
    LayerAdapter,
)
from bergson.utils.logger import get_logger
from bergson.utils.peft import set_peft_enabled


@dataclass
class HookCollectorBase(ContextDecorator, ABC):
    """
    Abstract base class for collectors that attach forward and backward hooks to model
    layers.

    Automatically discovers supported modules in the model, registers hooks during
    context entry, and provides lifecycle methods (setup/teardown) for subclasses to
    implement custom logic.

    Assumes module activation shape is [N, S, I] and activation gradient shape [N,S,O]
    where N=batch size, S=sequence length, I=input dimension, O=output dimension.

    Subclasses must implement:
        - setup(): Initialize state (buffers, dicts, etc.)
        - teardown(): Clean up and save results
        - forward_hook(): Process activations during forward pass
        - backward_hook(): Process gradients during backward pass
    """

    model: nn.Module
    """ The model to attach forward and backward hooks to. """

    filter_modules: str | None = None
    """If provided, a glob pattern to filter out modules from gradient collection.
    For example, "transformer.h.*.mlp.*" will exclude all MLP layers in a
    standard transformer architecture."""

    target_modules: set[str] | None = None
    """
    Set of module names to attach hooks to. Should consist only of supported modules
    (see LayerAdapter.supported_modules). If None, hooks are attached to all supported
    layers in the model.
    """

    processor: GradientProcessor = field(default_factory=GradientProcessor)
    """Configuration for processing and compressing gradients."""

    attention_cfgs: dict[str, AttentionConfig] = field(default_factory=dict)
    """
    Optional configuration specifying how to split up the attention module gradients
    into per-head gradients. See also bergson.config.AttentionConfig.
    """
    logger = get_logger("HookCollectorBase", level="INFO")

    def __post_init__(
        self,
    ):
        """Init, discover target modules, and call setup()."""
        self.rank = dist.get_rank() if dist.is_initialized() else 0
        self.world_size = dist.get_world_size() if dist.is_initialized() else 1

        self._fwd_hooks: list[RemovableHandle] = []
        self._bwd_hooks: list[RemovableHandle] = []

        # Discover target modules using the static method
        self.target_info = self.discover_targets(
            self.model,
            self.target_modules,
            self.processor.include_bias,
            self.filter_modules,
        )

        # Allow subclasses to perform custom initialization
        self.setup()

    @staticmethod
    def discover_targets(
        model: nn.Module,
        target_modules: set[str] | None = None,
        include_bias: bool = False,
        filter_modules: str | None = None,
    ) -> dict[str, tuple[torch.device, torch.Size, bool]]:
        """
        Discover target modules without instantiating a collector.

        This is useful when you need target_info early (e.g., to allocate buffers)
        before creating the actual collector instance.

        Args:
            model: The model to scan for supported layers, see
            LayerAdapter.supported_modules.
            target_modules: Optional set of module names to filter. If None, all
            supported layers are included.
            include_bias: Whether to track bias parameters for modules that have them.
            filter_modules: Optional glob pattern to exclude modules by name
            (e.g., "*.lm_head").

        Returns:
            Dictionary mapping module names to (device, weight_shape, has_bias) tuples.
        """
        target_info = {}
        for name, layer in model.named_modules():
            if not isinstance(layer, LayerAdapter.supported_modules):
                continue

            if target_modules is not None and name not in target_modules:
                continue

            if filter_modules and fnmatchcase(name, filter_modules):
                continue

            has_bias = getattr(layer, "bias", None) is not None and include_bias

            target_info[name] = (
                layer.weight.device,
                layer.weight.shape,
                has_bias,
            )
        return target_info

    @staticmethod
    def get_head_name(name: str, head_idx: int) -> str:
        """Get the name of an attention head with index `head_idx` in a
        module with name `name`."""
        return f"{name}.head_{head_idx}"

    @staticmethod
    def split_attention_heads(fn):
        """Decorator that splits attention module calls into per-head calls."""

        @functools.wraps(fn)
        def wrapper(self, module, g):
            name = module._name

            if name not in self.attention_cfgs:
                return fn(self, module, g)

            num_heads, head_size, head_dim = astuple(self.attention_cfgs[name])

            # Save state
            orig_name = module._name
            orig_out = getattr(module, LayerAdapter.out_attr(module))

            setattr(module, LayerAdapter.out_attr(module), head_size)

            for h in range(num_heads):
                module._name = self.get_head_name(name, h)
                try:
                    head_g = torch.narrow(g, head_dim, h * head_size, head_size)
                except Exception as e:
                    print(
                        f"Error processing gradient of shape {g.shape} for head {h}"
                        f" in module {name}. Provided head config may be incorrect. "
                        f"Head config: head dim {head_dim}, head size {head_size},"
                        f" num heads {num_heads}."
                    )
                    raise e
                fn(self, module, head_g)

            # Restore
            module._name = orig_name
            setattr(module, LayerAdapter.out_attr(module), orig_out)

        return wrapper

    def shapes(self) -> Mapping[str, torch.Size]:
        """Return the shapes of the gradients collected by this collector."""
        proj_shape = (
            torch.Size((p_dim, p_dim))
            if (p_dim := self.processor.projection_dim) is not None
            else None
        )

        shapes = {}
        for name, (_, target_shape, has_bias) in self.target_info.items():
            if name in self.attention_cfgs:
                attention_cfg = self.attention_cfgs[name]
                if proj_shape:
                    head_shape = proj_shape
                else:
                    # Mutate the attention module's shape to get the attention
                    # head shape
                    attention_shape = list(target_shape)
                    # - 2 because we're excluding the batch and sequence activation
                    # dimensions
                    attention_shape[attention_cfg.head_dim - 2] = (
                        attention_cfg.head_size
                    )
                    if has_bias:
                        attention_shape[-1] += 1
                    head_shape = torch.Size(attention_shape)

                shapes.update(
                    {
                        self.get_head_name(name, h): head_shape
                        for h in range(attention_cfg.num_heads)
                    }
                )
            else:
                if proj_shape:
                    shapes[name] = proj_shape
                else:
                    grad_shape = list(target_shape)
                    if has_bias:
                        grad_shape[-1] += 1
                    shapes[name] = torch.Size(grad_shape)

        return shapes

    def projection(
        self,
        name: str,
        m: int,
        n: int,
        side: Literal["left", "right"],
        device: torch.device,
        dtype: torch.dtype,
    ) -> Tensor:
        """Return the `side` projection matrix for parameter `name` of shape [m, n]."""
        key = (name, side, device)
        if key in self.processor._projection_matrices:
            return self.processor._projection_matrices[key]

        identifier = f"{name}/{side}"

        A = create_projection_matrix(
            identifier, m, n, dtype, device, self.processor.projection_type
        )
        self.processor._projection_matrices[key] = A
        return A

    def with_batch(self, valid_mask: Tensor | None = None) -> "HookCollectorBase":
        """
        Set the current batch indices and valid mask before entering the context.

        This allows hooks to access batch indices and valid mask during
        forward/backward passes.
        Usage:
            with collector.with_batch(indices, valid_mask):
                # forward/backward pass
                # hooks can access self._current_indices and self._current_valid_mask

        Args:
            indices: List of data indices in the current batch.
            valid_mask: Optional boolean tensor of shape [batch_size, seq_len]
                indicating which positions have valid labels for loss computation.

        Returns:
            self, for use as a context manager.
        """
        self._current_valid_mask = valid_mask
        return self

    def __enter__(self):
        """Register forward and backward hooks on all target modules."""
        for name in self.target_info:
            layer = self.model.get_submodule(name)

            # Store module name for use in hook callbacks
            layer._name = name  # type: ignore[attr-defined]
            layer._has_bias = self.target_info[name][2]  # type: ignore[attr-defined]

            # Register hooks
            fwd_hook = layer.register_forward_hook(self._process_input)
            self._fwd_hooks.append(fwd_hook)

            bwd_hook = layer.register_full_backward_hook(self._process_grad)
            self._bwd_hooks.append(bwd_hook)

        return self

    def _process_input(self, module: nn.Module, inp: tuple, _):
        """Internal forward hook that extracts input and delegates to subclass."""

        x = inp[0].detach()
        assert x.ndim == 3, f"Expected input of shape [N, S, I], got {x.shape}"

        self.forward_hook(module, x)

    def _process_grad(self, module: nn.Module, _, grad_out):
        """Internal backward hook that extracts gradient and delegates to subclass."""
        # Sanity checks
        assert isinstance(module, LayerAdapter.supported_modules), (
            f"Expected a module of type {LayerAdapter.supported_modules}, "
            f"got {type(module)}"
        )

        g = grad_out[0].detach()  # [N, S, O]

        self.backward_hook(module, g)

    def __exit__(self, exc_type, exc, tb):
        """Clean up hooks and allow subclass cleanup."""

        # Clean up temporary attributes
        for layer in self.model.modules():
            if hasattr(layer, "_inputs"):
                del layer._inputs
            if hasattr(layer, "_name"):
                del layer._name

        # Remove all registered hooks
        for h in self._fwd_hooks:
            h.remove()
        for h in self._bwd_hooks:
            h.remove()
        self._fwd_hooks.clear()
        self._bwd_hooks.clear()

        return False

    @abstractmethod
    def setup(self) -> None:
        """
        Called at the end of __post_init__.

        Override to perform custom initialization such as:
        - Allocating buffers or dictionaries
        - Loading pretrained weights or data
        - Initializing accumulators
        """
        pass

    @abstractmethod
    def teardown(self) -> None:
        """
        Called at the end.

        Override to perform custom cleanup such as:
        - Saving results to disk
        - Flushing buffers
        - Computing final statistics
        - Freeing resources
        """
        pass

    @abstractmethod
    def forward_hook(self, module: nn.Module, a: Float[Tensor, "N S I"]) -> None:
        """
        Process activations during the forward pass.

        Args:
            module: The module whose forward pass triggered this hook. The module name
                is available via module._name.
            a: Input activations of shape [N, S, I] where N=batch size, S=sequence
                length, I=input dimension.
        """
        pass

    @abstractmethod
    def backward_hook(self, module: nn.Module, g: Float[Tensor, "N S O"]) -> None:
        """
        Process gradients during the backward pass.

        Args:
            module: The module whose backward pass triggered this hook. The module name
                is available via module._name.
            g: Gradient with respect to module output, shape [N, S, O] where N=batch
                size, S=sequence length, O=output dimension.
        """
        pass

    @abstractmethod
    def process_batch(self, indices: list[int], **kwargs) -> None:
        """
        Process collected data for a batch. This is called after each
        forward/backward pass. See also CollectorComputer.run_with_collector_hooks.

        Args:
            indices: List of data indices in the current batch
            **kwargs: Additional batch-specific data (e.g., losses)
        """
        pass


class CollectorComputer:
    """
    Orchestrates gradient collection by running forward/backward passes over a dataset.

    Iterates through batches of data, computes losses, triggers backpropagation, and
    delegates gradient processing to the provided collector. Supports distributed
    training and optional profiling via PyTorch profiler via cfg.profile flag.
    """

    def __init__(
        self,
        model: PreTrainedModel,
        data: Dataset,
        *,
        collector: HookCollectorBase,
        batches: list[list[int]] | None = None,
        cfg: IndexConfig,
    ):
        """
        Initialize the CollectorComputer.

        Args:
            model: The model to collect gradients from.
            data: HuggingFace Dataset containing input_ids and optionally labels.
            collector: A HookCollectorBase instance that will process the gradients
            via hooks.
            batches: List of index lists defining how to batch the data. If None,
                defaults to batch size 1 (each sample processed individually).
            cfg: IndexConfig controlling all other hyperparameters.
        """
        # Model
        self.model = model
        self.device = model.device

        # Data
        self.data = data
        # Batch size one by default
        if batches is None:
            batches = [[idx] for idx in range(len(data))]
        self.batches = batches

        self.forward_backward = fwd_bwd_factory(cfg)

        # Collector
        self.collector = collector

        # Other
        self.cfg = cfg
        self.logger = get_logger(
            "CollectorComputer", level="DEBUG" if cfg.debug else "INFO"
        )

        # Distributed related
        self.rank = dist.get_rank() if dist.is_initialized() else 0
        self.world_size = dist.get_world_size() if dist.is_initialized() else 1

        self.logger.info("Computing with collector for target modules.")

    def _setup_profiler(self):
        """Set up profiler if profiling is enabled."""
        if not self.cfg.profile:
            return nullcontext()

        trace_handler = tensorboard_trace_handler(
            dir_name="profiler_logs", worker_name=f"rank_{self.rank}", use_gzip=True
        )
        my_schedule = schedule(wait=0, warmup=0, active=4, repeat=1)
        prof = profile(
            activities=[ProfilerActivity.CPU, ProfilerActivity.CUDA],
            on_trace_ready=trace_handler,
            schedule=my_schedule,
            record_shapes=True,
            with_stack=True,
            profile_memory=True,
            with_modules=True,
        )

        log_dir = "profiler_logs"
        os.makedirs(log_dir, exist_ok=True)

        return prof

    def run_with_collector_hooks(
        self,
        desc: Optional[str] = None,
    ):
        """
        Run the main computation loop over all batches.

        For each batch: computes forward pass, calculates loss, triggers backward pass
        (which invokes collector hooks), then calls collector.process_batch(). After
        all batches are processed, calls collector.teardown().

        Args:
            desc: Optional description string for the tqdm progress bar.
        """
        total_processed = torch.tensor(0, device=self.model.device)
        prof = self._setup_profiler()
        step = 0
        with prof:
            for indices in tqdm(
                self.batches,
                desc=f"Computing {desc}",
            ):
                batch = self.data[indices]

                # Compute padded tensors and valid_mask before entering context
                x, y, valid_mask = pad_and_tensor(
                    batch["input_ids"],
                    labels=batch.get("labels"),
                    device=self.model.device,
                )
                total_processed += valid_mask.sum()

                with (
                    self.collector.with_batch(valid_mask),
                    (
                        record_function(f"step_{step}")
                        if self.cfg.profile
                        else nullcontext()
                    ),
                ):
                    losses = self.forward_backward(self.model, x, y, batch)

                    # TODO: currently builder also calls torch.cuda.synchronize
                    torch.cuda.synchronize() if torch.cuda.is_available() else None

                if self.cfg.profile:
                    assert isinstance(prof, profile), "Profiler is not set up correctly"
                    prof.step()
                step += 1

                self.collector.process_batch(indices, losses=losses)

        self.collector.teardown()

        if dist.is_initialized():
            dist.all_reduce(total_processed, op=dist.ReduceOp.SUM)

        if self.rank == 0:
            torch.save(
                total_processed,
                os.path.join(self.cfg.partial_run_path, "total_processed.pt"),
            )
        self.logger.info(f"Total processed: {total_processed.item()}")


def fwd_bwd_factory(cfg: IndexConfig) -> Callable:
    """
    Create a forward/backward function based on the configuration.

    Args:
        cfg: IndexConfig that specifies:
            - cfg.loss_fn: Either "kl" for KL divergence (requires PEFT model) or
              any other value for cross-entropy loss.
            - cfg.loss_reduction: Either "mean" to average over tokens, or "sum" for
              summed loss.

    Returns:
        A callable fwd_bwd(model, x, y, batch) -> Tensor that performs a forward pass
        and backward pass, returning the per-sample losses.
        Args:
            model: The model to run forward/backward on.
            x: Padded input token ids tensor of shape [batch_size, seq_len].
            y: Padded label tensor of shape [batch_size, seq_len] with -100 for padding.
            batch: Original batch dict, used only for "advantage" if present.
        Returns a tensor of shape [batch_size] with one loss value per sample.
    """

    def fwd_bwd(model, x: Tensor, y: Tensor, batch: dict):
        logits = model(x).logits[:, :-1]
        masks = y[:, 1:] != -100
        denoms = (
            masks.sum(dim=1, dtype=model.dtype) if cfg.loss_reduction == "mean" else 1.0
        )

        if cfg.loss_fn == "kl":
            with torch.inference_mode():
                set_peft_enabled(model, False)
                ref_lps = torch.log_softmax(model(x).logits[:, :-1], dim=-1)
                set_peft_enabled(model, True)

            ft_lps = torch.log_softmax(logits, dim=-1)

            # Compute average KL across all unmasked tokens
            kls = torch.sum(ft_lps.exp() * (ft_lps - ref_lps), dim=-1)
            losses = torch.sum(kls * masks, dim=-1) / denoms
            if "advantage" in batch:
                losses *= torch.tensor(batch["advantage"], device=losses.device)

        else:
            losses = F.cross_entropy(
                logits.reshape(-1, logits.size(-1)),
                y[:, 1:].flatten(),
                reduction="none",
                label_smoothing=cfg.label_smoothing,
            ).reshape_as(y[:, 1:])
            losses = losses.sum(1) / denoms
            if "advantage" in batch:
                losses *= torch.tensor(batch["advantage"], device=losses.device)

        losses.sum().backward()
        model.zero_grad()

        return losses

    return fwd_bwd


def fwd_bwd_hessian_factory(
    index_cfg: IndexConfig, hessian_cfg: HessianConfig
) -> Callable:
    def fwd_bwd_hessian(model, x: Tensor, y: Tensor, batch: dict):
        logits = model(x).logits[:, :-1]
        masks = y[:, 1:] != -100
        denoms = (
            masks.sum(dim=1, dtype=model.dtype)
            if index_cfg.loss_reduction == "mean"
            else 1.0
        )
        if hessian_cfg.use_dataset_labels:
            losses = F.cross_entropy(
                logits.reshape(-1, logits.size(-1)),
                y[:, 1:].flatten(),
                reduction="none",
            ).reshape_as(y[:, 1:])
            losses = losses.sum(1) / denoms
        else:
            with torch.no_grad():
                probs = F.softmax(logits, dim=-1)
                sampled_tokens = torch.multinomial(
                    probs.reshape(-1, probs.size(-1)),
                    num_samples=1,
                    replacement=True,
                ).reshape_as(y[:, 1:])
            losses = F.cross_entropy(
                logits.reshape(-1, logits.size(-1)),
                sampled_tokens.flatten(),
                reduction="none",
            ).reshape_as(y[:, 1:])
            losses = losses.sum(1) / denoms

        losses.sum().backward()
        model.zero_grad()

        return losses

    return fwd_bwd_hessian


def create_projection_matrix(
    identifier: str,
    m: int,
    n: int,
    dtype: torch.dtype,
    device: torch.device,
    projection_type: Literal["normal", "rademacher"] = "normal",
) -> Tensor:
    """Create a projection matrix deterministically based on identifier and side."""
    # Seed the PRNG with the name of the layer and what "side" we are projecting
    message = bytes(identifier, "utf-8")
    digest = hashlib.md5(message).digest()
    seed = int.from_bytes(digest, byteorder="big") % (2**63 - 1)

    if projection_type == "normal":
        prng = torch.Generator(device).manual_seed(seed)
        A = torch.randn(m, n, device=device, dtype=dtype, generator=prng)
    elif projection_type == "rademacher":
        numpy_rng = np.random.Generator(np.random.PCG64(seed))
        random_bytes = numpy_rng.bytes((m * n + 7) // 8)
        random_bytes = np.frombuffer(random_bytes, dtype=np.uint8)
        A = np.unpackbits(random_bytes)[: m * n].reshape((m, n))
        A = torch.from_numpy(A).to(device, dtype=dtype)
        A = A.add_(-0.5).mul_(2)
    else:
        raise ValueError(f"Unknown projection type: {projection_type}")
    A /= A.norm(dim=1, keepdim=True)
    return A



from contextlib import contextmanager

import torch
from torch import Tensor, nn
from transformers import PreTrainedModel

from ..gradients import Normalizer


class FiniteDiff:
    def __init__(
        self,
        model: nn.Module,
        *,
        normalizers: dict[str, Normalizer] | None = None,
    ):
        self.model = model
        self.normalizers = normalizers or {}
        self.params: dict[str, Tensor] = {}

    def store(self, step_size: float):
        """Compute and store finite differences for the model parameters.

        This method assumes that you've just called .backward() on some loss function,
        so the model's parameters have gradients. We reuse these gradient buffers and
        set the .grad attributes to None to avoid unnecessary memory usage.
        """
        num_normalized = 0

        for name, param in self.model.named_parameters():
            if (g := param.grad) is None:
                continue

            ## Normalize the gradient if needed
            mod_name = name.removesuffix(".weight")
            if (n := self.normalizers.get(mod_name)) is not None:
                n.normalize_(g)
                num_normalized += 1

            # Compute the finite difference in-place *in the grad buffer*
            torch.add(param.data, g, alpha=step_size, out=g)

            # Stash the grad buffer and set p.grad to None
            self.params[name] = g
            param.grad = None

        # Sanity check to make sure we actually use the normalizers
        if num_normalized != len(self.normalizers):
            raise ValueError(
                f"Expected {len(self.normalizers)} normalized gradients, "
                f"but got {num_normalized}."
            )

    def swap(self):
        """Swap the original and updated parameters."""
        for name, param in self.model.named_parameters():
            if (buf := self.params.get(name)) is None:
                continue

            # Perform efficient XOR swap to avoid copying
            assert buf.shape == param.shape
            a_i, b_i = buf.view(torch.uint8), param.data.view(torch.uint8)
            a_i.bitwise_xor_(b_i)  # a = a ^ b
            b_i.bitwise_xor_(a_i)  # b = (a ^ b) ^ b = a
            a_i.bitwise_xor_(b_i)  # a = (a ^ b) ^ a = b

    @contextmanager
    def apply(self):
        """Context manager to apply finite differences to the model."""
        self.swap()
        try:
            yield self
        finally:
            self.swap()

    def clear(self):
        """
        Clear the stored finite differences.
        """
        self.params.clear()


def compute_effect(
    model: PreTrainedModel,
    x: Tensor,
    y: Tensor,
    *,
    eps: float | None = None,
    normalizers: dict[str, Normalizer] | None = None,
):
    """Compute the normalized effect on `y` of perturbing the model along `x` grads."""
    fd = FiniteDiff(
        model.base_model,
        normalizers=normalizers,
    )
    eps = eps or torch.finfo(model.dtype).eps

    with torch.no_grad():
        loss1 = model(y, labels=y).loss

    model(x, labels=x).loss.backward()
    fd.store(eps)
    model.zero_grad()

    with fd.apply(), torch.no_grad():
        loss2 = model(y, labels=y).loss

    fd.clear()

    # Divide by the number of parameters because for small eps, the diff is an inner
    # product of gradients, whose scale will be proportional to the number of
    # parameters, especially if the two inputs are similar.
    N = sum(p.numel() for p in model.parameters())
    return (loss2 - loss1) / (eps * N)



import math
from pathlib import Path

import numpy as np
import pytest
import torch
from transformers import AutoConfig, AutoModelForCausalLM

from bergson.collector.gradient_collectors import GradientCollector
from bergson.config import IndexConfig
from bergson.data import create_index, load_gradients


@pytest.mark.skipif(not torch.cuda.is_available(), reason="CUDA not available")
def test_large_gradients_build(tmp_path: Path, dataset):
    # Create index for uncompressed gradients from a large model.
    config = AutoConfig.from_pretrained(
        "EleutherAI/pythia-1.4b", trust_remote_code=True
    )
    model = AutoModelForCausalLM.from_config(config)
    collector = GradientCollector(
        model=model.base_model, data=dataset, cfg=IndexConfig(run_path=str(tmp_path))
    )
    grad_sizes = {name: math.prod(s) for name, s in collector.shapes().items()}

    create_index(
        tmp_path,
        num_grads=len(dataset),
        grad_sizes=grad_sizes,
        dtype=np.float32,
        with_structure=False,
    )

    # Load a large gradient index without structure.
    load_gradients(tmp_path, structured=False)

    with pytest.raises(ValueError):
        # Max item size exceeded.
        load_gradients(tmp_path, structured=True)



import os
from dataclasses import dataclass

import torch
import torch.distributed as dist
import torch.nn as nn
from safetensors.torch import save_file
from torch import Tensor

from bergson.collector.collector import HookCollectorBase
from bergson.hessians.sharded_computation import ShardedMul
from bergson.utils.utils import assert_type


@dataclass(kw_only=True)
class TraceCovarianceCollector(HookCollectorBase):
    """
    Collects activation and gradient covariances for TKFAC.

    Computes:
        A_tcov = sum over batches of (X^T @ X)  for activations * trace(S_cov)
        S_tcov = sum over batches of (G^T @ G)  for gradients * trace(A_cov)
        trace= sum over batches of trace(X^T @ X) * trace(G^T @ G)

    where X is input activations [N*S, I] and G is output gradients [N*S, O].
    """

    dtype: torch.dtype
    path: str

    def setup(self) -> None:
        """Initialize covariance storage dictionaries."""
        self.A_tcov_dict = {}
        self.S_tcov_dict = {}
        self.trace_dict = {}
        self.shard_computer = ShardedMul()
        # Initialize sharded covariance matrices for ALL modules in target_info
        self.shard_computer._init_covariance_dict(
            activation_covariance_dict=self.A_tcov_dict,
            gradient_covariance_dict=self.S_tcov_dict,
            dtype=self.dtype,
            target_info=self.target_info,
        )

    def forward_hook(self, module: nn.Module, a: Tensor) -> None:
        """Compute activation covariance: A^T @ A."""

        mask = self._current_valid_mask
        assert mask is not None, "Valid mask not set for forward hook."

        # a: [N, S, I], valid_masks: [N, S] -> select valid positions
        a_bi = a[mask]  # [num_valid, I]

        module._inputs = a_bi

    def backward_hook(self, module: nn.Module, g: Tensor) -> None:
        """Compute gradient covariance: G^T @ G."""
        name = assert_type(str, module._name)
        S_tcov_po = self.S_tcov_dict[name]
        A_tcov_ki = self.A_tcov_dict[name]
        mask = self._current_valid_mask

        # g: [N, S, O], mask: [N, S] -> select valid positions
        g_bo = g[mask]  # [num_valid, O]
        a_bi = module._inputs
        del module._inputs
        assert isinstance(a_bi, Tensor)

        # Compute local covariances
        local_update_oo = g_bo.mT @ g_bo
        local_update_ii = a_bi.mT @ a_bi

        # Compute traces
        trace_oo = (local_update_oo.diagonal()).sum()
        trace_ii = (local_update_ii.diagonal()).sum()

        local_update_oo = local_update_oo * trace_ii
        local_update_ii = local_update_ii * trace_oo

        # All-reduce across ranks
        if dist.is_initialized():
            dist.all_reduce(local_update_oo, op=dist.ReduceOp.SUM)
            dist.all_reduce(local_update_ii, op=dist.ReduceOp.SUM)

        # Extract our shard
        start_row_grad = self.rank * S_tcov_po.shape[0]
        end_row_grad = (self.rank + 1) * S_tcov_po.shape[0]
        update_slice_po = local_update_oo[start_row_grad:end_row_grad, :]

        start_row_act = self.rank * A_tcov_ki.shape[0]
        end_row_act = (self.rank + 1) * A_tcov_ki.shape[0]
        update_slice_ki = local_update_ii[start_row_act:end_row_act, :]

        # Accumulate
        S_tcov_po.add_(update_slice_po)
        A_tcov_ki.add_(update_slice_ki)

    def process_batch(self, indices: list[int], **kwargs) -> None:
        """No per-batch processing needed for covariance collection."""
        pass

    def teardown(self) -> None:
        """Save covariance matrices to disk."""
        activation_path = os.path.join(self.path, "activation_sharded")
        gradient_path = os.path.join(self.path, "gradient_sharded")

        os.makedirs(activation_path, exist_ok=True)
        os.makedirs(gradient_path, exist_ok=True)
        self.logger.info(
            f"Saving sharded covariance matrices to {activation_path} "
            f"and {gradient_path}"
        )
        # Save sharded covariance matrices
        save_file(
            self.A_tcov_dict,
            os.path.join(activation_path, f"shard_{self.rank}.safetensors"),
        )
        save_file(
            self.S_tcov_dict,
            os.path.join(gradient_path, f"shard_{self.rank}.safetensors"),
        )
        # TODO: Multiply by trace and save trace_dict



__version__ = "0.5.1"

from .collection import collect_gradients
from .collector.gradient_collectors import GradientCollector
from .config import (
    AttentionConfig,
    DataConfig,
    IndexConfig,
    QueryConfig,
    ReduceConfig,
    ScoreConfig,
)
from .data import load_gradient_dataset, load_gradients
from .gradients import GradientProcessor
from .normalizer.fit_normalizers import fit_normalizers
from .query.attributor import Attributor
from .query.faiss_index import FaissConfig
from .score.scorer import Scorer
from .utils.gradcheck import FiniteDiff

__all__ = [
    "collect_gradients",
    "load_gradients",
    "load_gradient_dataset",
    "fit_normalizers",
    "Attributor",
    "FaissConfig",
    "FiniteDiff",
    "GradientProcessor",
    "GradientCollector",
    "IndexConfig",
    "DataConfig",
    "AttentionConfig",
    "Scorer",
    "ScoreConfig",
    "ReduceConfig",
    "QueryConfig",
]



import math
import os
from functools import wraps
from itertools import chain
from pathlib import Path
from typing import Sized

import numpy as np
import torch
import torch.distributed as dist
from datasets import Dataset
from peft import PeftModel
from torch import Tensor
from torch.utils.data import DataLoader
from transformers.trainer import Trainer
from transformers.trainer_callback import TrainerCallback, TrainerControl, TrainerState
from transformers.training_args import TrainingArguments

from bergson import AttentionConfig, GradientProcessor
from bergson.collector.gradient_collectors import StreamingGradientCollector
from bergson.data import create_index
from bergson.gradients import AdafactorNormalizer, AdamNormalizer
from bergson.utils.peft import detect_peft_modules
from bergson.utils.utils import convert_dtype_to_torch


class GradientCollectorCallback(TrainerCallback):
    """Callback that collects gradients from the model during training.
    Implements the same functionality as `collect_gradients`."""

    def __init__(
        self,
        path: Path,
        attention_cfgs: dict[str, AttentionConfig] = {},
        projection_dim: int = 16,
        include_bias: bool = False,
        dtype: np.dtype = np.dtype(np.float16),
        accumulate_grads: bool = False,
        use_optimizer_state: bool = True,
        track_order: bool = False,
    ):
        """
        Args:
            path: The path to save the gradients
            projection_dim: The dimension to project the gradients onto
            include_bias: Whether to append bias gradients when present on a module
            dtype: The dtype of the on-disk gradient store
            accumulate_grads: Whether to take the sum of the gradients
                of the same example across epochs. If `False`, the
                gradients for each epoch are stored separately.
            use_optimizer_state: Whether to use the optimizer state to
                normalize the gradients. If `False`, no normalization is
                applied.
            track_order: Whether to record the shuffled order of training data.
        attention_cfgs: Information used to split matrix-valued parameters into
            per-head matrices before down projection.
        """
        super().__init__()

        # Initialized in on_train_begin when we learn what the model is
        self.collector: StreamingGradientCollector
        self.grad_sizes = {}

        self.attention_cfgs = attention_cfgs
        self.accumulate_grads = accumulate_grads
        self.dtype = dtype
        self.path = path
        self.projection_dim = projection_dim
        self.include_bias = include_bias
        self.use_optimizer_state = use_optimizer_state
        self.order: list[dict] | None = [] if track_order else None

        self.eval_grad_buffers: dict[str, np.memmap] = {}
        self.eval_step_idxs: dict[str, int] = {}
        self.eval_indices: dict[str, list[int]] = {}

        self.mod_grads = {}
        self.batch_indices: Tensor | None = None

        self.torch_dtype = convert_dtype_to_torch(self.dtype)

    def write_grads(self, grad_buffer: np.memmap):
        torch.cuda.synchronize()
        for layer_name, g in self.collector.mod_grads.items():
            grad_buffer[layer_name][self.batch_indices, :] = g.numpy()

        self.collector.mod_grads.clear()

    def on_train_begin(
        self,
        args: TrainingArguments,
        state: TrainerState,
        control: TrainerControl,
        *,
        model: torch.nn.Module,
        **kwargs,
    ):
        if not hasattr(args, "__gradient_collection_enabled__"):
            raise RuntimeError(
                "Gradient collection is not enabled. Please enable it by "
                "calling bergson.prepare_gradient_collection on the trainer."
            )

        if isinstance(model, PeftModel):
            reshape_to_square = True
            target_modules = detect_peft_modules(model)  # type: ignore
        else:
            reshape_to_square = False
            target_modules = None

        self.collector = StreamingGradientCollector(
            model=getattr(model, "base_model", model),
            processor=GradientProcessor(
                {},
                projection_dim=self.projection_dim or None,
                reshape_to_square=reshape_to_square,
                include_bias=self.include_bias,
            ),
            target_modules=target_modules,
            attention_cfgs=self.attention_cfgs,
            dtype=self.torch_dtype,
        )
        self.grad_sizes = {
            name: math.prod(s) for name, s in self.collector.shapes().items()
        }

        # Record forward and backward hooks
        self.collector.__enter__()
        self.fwd_handle = model.register_forward_pre_hook(
            self.on_forward_begin,
            with_kwargs=True,
        )

    def on_epoch_begin(
        self,
        args: TrainingArguments,
        state: TrainerState,
        control: TrainerControl,
        *,
        eval_dataloader: DataLoader | dict[str, DataLoader],
        train_dataloader: DataLoader,
        **kwargs,
    ):
        epoch = int(state.epoch or 0)

        if self.accumulate_grads and epoch > 0:
            return

        epoch_suffix = "" if self.accumulate_grads else f"/epoch_{epoch}"

        ds = train_dataloader.dataset
        if not isinstance(ds, Sized):
            raise ValueError("Dataset must be sized for gradient collection")

        self.train_grad_buffer = create_index(
            self.path / ("train" + epoch_suffix),
            num_grads=len(ds),
            grad_sizes=self.grad_sizes,
            dtype=self.dtype,
        )
        self.train_step_idx = 0

        # Set up the gradient buffers for the evaluation datasets
        if eval_dataloader is None:
            return
        elif isinstance(eval_dataloader, dict):
            eval_datasets = eval_dataloader
        else:
            eval_datasets = {"eval": eval_dataloader}

        for dataset_name, dataloader in eval_datasets.items():
            self.eval_grad_buffers[dataset_name] = create_index(
                self.path / (dataset_name + epoch_suffix),
                num_grads=len(dataloader),
                grad_sizes=self.grad_sizes,
                dtype=self.dtype,
            )
            self.eval_step_idxs[dataset_name] = 0
            self.eval_indices[dataset_name] = [
                i.item() for batch in dataloader for i in batch["_idx"]
            ]

    def on_epoch_end(
        self,
        args: TrainingArguments,
        state: TrainerState,
        control: TrainerControl,
        **kwargs,
    ):
        rank = dist.get_rank() if dist.is_initialized() else 0
        if rank == 0:
            epoch = int(state.epoch or 0) - 1
            epoch_suffix = "" if self.accumulate_grads else f"/epoch_{epoch}"
            path = self.path / ("train" + epoch_suffix)

            assert self.collector is not None
            self.collector.processor.save(path)

        # Ensure the gradients are written to disk
        self.train_grad_buffer.flush()
        for eval_grad_buffer in self.eval_grad_buffers.values():
            eval_grad_buffer.flush()

    def on_forward_begin(self, _: torch.nn.Module, args, kwargs: dict):
        # Record the original indices of this batch
        self.batch_indices = kwargs.pop("_idx").to("cpu", non_blocking=True)
        return args, kwargs

    def on_module_backward(self, name: str, g: Tensor):
        lo = torch.finfo(self.torch_dtype).min
        hi = torch.finfo(self.torch_dtype).max
        g = g.flatten(1).clamp_(lo, hi)

        # Asynchronously move the gradient to CPU and convert to fp16
        self.mod_grads[name] = g.to(
            device="cpu", dtype=self.torch_dtype, non_blocking=True
        )

    def on_substep_end(
        self,
        args: TrainingArguments,
        state: TrainerState,
        control: TrainerControl,
        **kwargs,
    ):
        """Called at the end of each training step.
        If using gradient accumulation, one training step might take several inputs."""
        self.write_grads(self.train_grad_buffer)

    def on_step_end(
        self,
        args: TrainingArguments,
        state: TrainerState,
        control: TrainerControl,
        *,
        model: torch.nn.Module,
        optimizer: torch.optim.Optimizer,
        **kwargs,
    ):
        self.on_substep_end(args, state, control)
        print("Step end")

        # Record training order if enabled
        if self.order is not None:
            assert (
                self.batch_indices is not None
            ), "Batch indices are not available for training order tracking"

            epoch = int(state.epoch or 0)
            global_step = state.global_step

            self.order.extend(
                {
                    "_idx": int(idx),
                    # global_step is 1-indexed.
                    "global_step": global_step,
                    "epoch": epoch,
                }
                for idx in self.batch_indices.tolist()
            )

        # We can skip all this if we're not using the optimizer state
        if not self.use_optimizer_state:
            return

        # The optimizer doesn't actually know the names of the parameters
        model = getattr(model, "base_model", model)
        param_to_name = {
            param: name
            for name, param in model.named_parameters()
            if param.requires_grad
        }
        normalizers: dict[str, AdafactorNormalizer] = {}

        assert self.collector is not None
        proc = self.collector.processor
        proc.normalizers = {}

        # Read normalizers off of the optimizer state. We need to figure out
        # what type of optimizer this is first.
        for group in optimizer.param_groups:
            lr_sqrt = group["lr"] ** 0.5

            for param in group["params"]:
                name = param_to_name[param].removesuffix(".weight")
                if name not in self.collector.target_info:
                    continue

                p_state = optimizer.state[param]

                # Adam-like optimizer
                if (eas := p_state.get("exp_avg_sq")) is not None:
                    norm = AdamNormalizer(eas).to_adafactor()

                # Adafactor-like optimizer
                elif (vr := p_state.get("exp_avg_sq_row")) is not None:
                    vc = p_state.get("exp_avg_sq_col")
                    norm = AdafactorNormalizer(vr, vc)
                else:
                    continue

                # Scale the gradient by the current learning rate. It's factorized
                # so we multiply each factor by the square root of the LR.
                norm.row *= lr_sqrt
                norm.col *= lr_sqrt
                normalizers[name] = norm

        proc.normalizers = normalizers

    def on_prediction_step(self, args, state, control, **kwargs):
        dataset_name = kwargs["inputs"]["dataset_name"]
        self.write_grads(self.eval_grad_buffers[dataset_name])

    def on_train_end(
        self,
        args: TrainingArguments,
        state: TrainerState,
        control: TrainerControl,
        **kwargs,
    ):
        assert self.collector is not None
        self.collector.__exit__(None, None, None)
        self.fwd_handle.remove()

        if self.order is not None:
            self._save_order()

    def _save_order(self):
        """Save the training order to disk, handling distributed training."""
        assert self.order is not None
        os.makedirs(self.path, exist_ok=True)

        if dist.is_initialized():
            # Gather training order from all processes
            all_orders = [None] * dist.get_world_size()
            dist.all_gather_object(all_orders, self.order)

            # Only rank 0 saves the merged data
            if dist.get_rank() == 0:
                merged_order = list(
                    chain.from_iterable(
                        order for order in all_orders if order is not None
                    )
                )
                dataset = Dataset.from_list(merged_order)
                dataset.save_to_disk(os.path.join(self.path, "order.hf"))

        else:
            dataset = Dataset.from_list(self.order)
            dataset.save_to_disk(os.path.join(self.path, "order.hf"))


def prepare_for_gradient_collection(trainer: Trainer):
    """Mutate the trainer and its datasets in-place to expose the datasets'
    indices to the gradient collector callback."""
    # Add indices to the training dataset
    trainer.train_dataset = trainer.train_dataset.map(  # type: ignore
        lambda ex, idx: {"_idx": idx}, with_indices=True
    )

    # Add indices to the evaluation dataset/s if they exist
    if trainer.eval_dataset is not None:
        if isinstance(trainer.eval_dataset, dict):
            for eval_name, dataset in trainer.eval_dataset.items():
                trainer.eval_dataset[eval_name] = dataset.map(  # type: ignore
                    lambda ex, idx: {"_idx": idx}, with_indices=True
                )
        else:
            trainer.eval_dataset = trainer.eval_dataset.map(  # type: ignore
                lambda ex, idx: {"_idx": idx}, with_indices=True
            )

    trainer._set_signature_columns_if_needed()
    trainer._signature_columns.append("_idx")  # type: ignore

    if trainer.data_collator:
        original_collator = trainer.data_collator

        @wraps(original_collator)  # type: ignore
        def wrapped_collator(features):
            batch = original_collator(features)
            batch.setdefault("_idx", torch.tensor([f["_idx"] for f in features]))
            return batch

        trainer.data_collator = wrapped_collator

    trainer.args.__gradient_collection_enabled__ = True  # type: ignore

    return trainer



import torch
import torch.distributed as dist
from jaxtyping import Float
from torch import Tensor


class ShardedMul:
    def __init__(
        self,
    ):
        self.dist = dist.is_initialized()

        self.rank = dist.get_rank() if self.dist else 0
        self.world_size = dist.get_world_size() if self.dist else 1
        self.device = torch.device(
            f"cuda:{self.rank}" if torch.cuda.is_available() else "cpu"
        )

    def _init_covariance_dict(
        self,
        activation_covariance_dict: dict,
        gradient_covariance_dict: dict,
        dtype: torch.dtype,
        target_info: dict[str, tuple[torch.device, torch.Size, bool]],
    ):
        """Initialize the covariance matrices for activations and gradients."""

        for name, (device, weight_shape, has_bias) in target_info.items():
            # Activation covariance A^T A has shape [in_dim, in_dim]
            in_dim = weight_shape[1]
            shard_in_dim = in_dim if not self.dist else in_dim // self.world_size
            activation_covariance_dict[name] = torch.zeros(
                (shard_in_dim, in_dim), device=self.device, dtype=dtype
            )

            # Gradient covariance G^T G has shape [out_dim, out_dim]
            out_dim = weight_shape[0]
            shard_out_dim = out_dim if not self.dist else out_dim // self.world_size
            gradient_covariance_dict[name] = torch.zeros(
                (shard_out_dim, out_dim), device=self.device, dtype=dtype
            )

    def _matmul(
        self,
        vector_nsa: Float[Tensor, "n s a"],
        matrix_cb: Float[Tensor, "c b"],
    ) -> Float[Tensor, "n s b"]:
        """Vector-matrix multiplication.
        - If not distributed, this does usual multiplication with a=c.
        - If distributed, assumes that c=a/world_size and does sharded multiplication.
        """

        assert vector_nsa.shape[2] == matrix_cb.shape[0] * self.world_size, (
            f"Vector shape {vector_nsa.shape} not compatible with matrix shape "
            f"{matrix_cb.shape} and world_size {self.world_size}"
        )

        if not self.dist:
            result_nsb = torch.einsum("n s c, c b-> n s b", vector_nsa, matrix_cb)

        else:
            result_nsb = self._sharded_matmul(vector_nsa, matrix_cb)

        return result_nsb

    def _transpose_matmul(
        self,
        vector_nsa: Float[Tensor, "n s a"],
        matrix_cb: Float[Tensor, "c b"],
    ) -> Float[Tensor, "n s b"]:
        if not self.dist:
            result_nsb = torch.einsum("n s c, b c -> n s b", vector_nsa, matrix_cb)
        else:
            result_nsb = self._sharded_transpose_matmul(vector_nsa, matrix_cb)
        return result_nsb

    def _hadamard(
        self,
        matrix_noi: Float[Tensor, "n o i"],
        lambda_ci: Float[Tensor, "c i"],
        lambda_damp_factor: float = 0.1,
    ):
        if not self.dist:
            global_lambda_mean = lambda_ci.mean()
            inverse_lambda = (
                lambda_ci + lambda_damp_factor * global_lambda_mean
            ).reciprocal()
            matrix_noi.mul_(inverse_lambda)
        else:
            self._sharded_hadamard(matrix_noi, lambda_ci, lambda_damp_factor)

    def _sharded_matmul(
        self,
        vector_nsa: Float[Tensor, "n s a"],
        matrix_cb: Float[Tensor, "c b"],
    ) -> Float[Tensor, "n s b"]:
        """
        Sharded matrix multiplication for distributed training.
        Assumes that c=a/world_size.
        vector: [n, s, a]
        matrix_shard: [a/world_size, b]
        Returns: [n, s, b]
        """
        # Split the vector into shards
        vector_shards_wnsc = torch.chunk(
            vector_nsa, self.world_size, dim=-1
        )  # (w, n, s, a/w)
        n, s, b = vector_nsa.shape[0], vector_nsa.shape[1], matrix_cb.shape[1]

        result_nsb = torch.zeros(
            (n, s, b),
            device=vector_nsa.device,
            dtype=vector_nsa.dtype,
        )

        for rank_index in range(self.world_size):
            if rank_index == self.rank:
                shard_cb = matrix_cb
            else:
                shard_cb = torch.zeros_like(matrix_cb)

            dist.broadcast(shard_cb, src=rank_index)
            result_nsb += torch.einsum(
                "n s c, c b-> n s b", vector_shards_wnsc[rank_index], shard_cb
            )  # [B, c]
            if self.rank != rank_index:
                del shard_cb

        return result_nsb

    def _sharded_hadamard(
        self,
        matrix_noi: Float[Tensor, "n o i"],
        lambda_ci: Float[Tensor, "c i"],
        lambda_damp_factor: float = 0.1,
    ):
        """
        Sharded in-place element-wise multiplication for distributed training.
        gradients: [n, o, i]
        matrix_shard: [c, i] where c=o/world_size

        """

        global_lambda_mean = lambda_ci.mean()

        dist.all_reduce(global_lambda_mean, op=dist.ReduceOp.SUM)
        global_lambda_mean /= self.world_size

        for rank_index in range(self.world_size):
            if rank_index == self.rank:
                shard_ci = lambda_ci
            else:
                shard_ci = torch.zeros_like(lambda_ci)

            dist.broadcast(shard_ci, src=rank_index)

            start_row = rank_index * shard_ci.shape[0]
            end_row = (rank_index + 1) * shard_ci.shape[0]
            inverse_lambda = (
                shard_ci + lambda_damp_factor * global_lambda_mean
            ).reciprocal()

            matrix_noi[:, start_row:end_row, :].mul_(inverse_lambda)

            if self.rank != rank_index:
                del shard_ci

    def _sharded_transpose_matmul(
        self,
        matrix_noi: Float[Tensor, "n o i"],
        matrix_bc: Float[Tensor, "b c"],
    ):
        """
        Sharded matrix multiplication for distributed training.
        Assumes that c=i/world_size if left or o/world_size if right.
        gradients: [n, o, i]
        matrix_shard: [c, b] where b=i if left or b=o if right
        Returns: [n, o, c*w] if left or [n, c*w, i] if right
        """

        x, y = (matrix_noi.shape[1], matrix_bc.shape[0] * self.world_size)

        result_nxy = torch.zeros(
            matrix_noi.shape[0], x, y, device=matrix_noi.device, dtype=matrix_noi.dtype
        )

        for rank_index in range(self.world_size):
            if rank_index == self.rank:
                shard_bc = matrix_bc
            else:
                shard_bc = torch.zeros_like(matrix_bc)
            dist.broadcast(shard_bc, src=rank_index)

            shard_size = shard_bc.shape[0]
            start_row = rank_index * shard_size
            end_row = (rank_index + 1) * shard_size

            result_nxy[:, :, start_row:end_row].copy_(
                torch.einsum("n o i, c i -> n o c", matrix_noi, shard_bc)
            )

            if self.rank != rank_index:
                del shard_bc

        return result_nxy



import torch
import torch.distributed as dist

from bergson.gradients import GradientProcessor


def process_preconditioners(
    processor: GradientProcessor,
    preconditioners: dict[str, torch.Tensor],
    len_data: int,
    grad_sizes: dict[str, int],
    rank: int,
):
    """
    Aggregate preconditioners across ranks and compute their eigen decomposition
    distributed across all ranks.
    """
    preconditioners_eigen = {}

    device = next(iter(preconditioners.values())).device
    dtype = next(iter(preconditioners.values())).dtype

    if rank == 0:
        print("Saving preconditioners...")

    for name, prec in preconditioners.items():
        preconditioners[name] = (prec / len_data).cpu()

    if rank == 0:
        print("Computing preconditioner eigen decompositions...")

    for name in preconditioners.keys():
        prec = preconditioners[name].to(dtype=torch.float64, device=device)
        eigvals, eigvecs = torch.linalg.eigh(prec)
        preconditioners_eigen[name] = (
            eigvals.to(dtype=dtype).contiguous().cpu(),
            eigvecs.to(dtype=dtype).contiguous().cpu(),
        )

    if not dist.is_initialized():
        processor.preconditioners = preconditioners
        processor.preconditioners_eigen = preconditioners_eigen
        return

    if rank == 0:
        print("Gathering preconditioners...")

    cpu_group: dist.ProcessGroup = dist.new_group(backend="gloo")  # type: ignore[assignment]

    for name, grad_size in grad_sizes.items():
        if name in preconditioners:
            local_prec = preconditioners[name]
            del preconditioners[name]
        else:
            local_prec = torch.zeros([grad_size, grad_size], dtype=dtype, device="cpu")

        dist.reduce(local_prec, dst=0, op=dist.ReduceOp.SUM, group=cpu_group)

        if rank == 0:
            preconditioners[name] = local_prec

    if rank == 0:
        processor.preconditioners = preconditioners

        print("Gathering eigen decompositions...")

    for name, grad_size in grad_sizes.items():
        prec_size = torch.Size([grad_size, grad_size])
        if name not in preconditioners_eigen:
            eigval = torch.zeros(prec_size[0], dtype=dtype)
            eigvec = torch.zeros(prec_size, dtype=dtype)
        else:
            eigval, eigvec = preconditioners_eigen[name]

        dist.reduce(eigval, dst=0, op=dist.ReduceOp.SUM, group=cpu_group)
        dist.reduce(eigvec, dst=0, op=dist.ReduceOp.SUM, group=cpu_group)

        if rank == 0:
            preconditioners_eigen[name] = (eigval, eigvec)

    if rank == 0:
        processor.preconditioners_eigen = preconditioners_eigen



import subprocess
from pathlib import Path

import pytest
import torch

from bergson.config import IndexConfig
from bergson.data import load_gradient_dataset


@pytest.mark.skipif(not torch.cuda.is_available(), reason="CUDA not available")
def test_reduce_e2e(tmp_path: Path):
    result = subprocess.run(
        [
            "python",
            "-m",
            "bergson",
            "reduce",
            "test_reduce_e2e",
            "--model",
            "EleutherAI/pythia-14m",
            "--dataset",
            "NeelNanda/pile-10k",
            "--split",
            "train[:100]",
            "--truncation",
            "--method",
            "mean",
            "--unit_normalize",
            "--skip_preconditioners",
            "--token_batch_size",
            "1024",
        ],
        cwd=tmp_path,
        capture_output=True,
        # Get strings instead of bytes
        text=True,
    )

    assert (
        "error" not in result.stderr.lower()
    ), f"Error found in stderr: {result.stderr}"

    # Load the gradient index
    index_cfg = IndexConfig(run_path=str(tmp_path / "test_reduce_e2e"))
    ds = load_gradient_dataset(Path(index_cfg.run_path), structured=False)
    assert len(ds) == 1

    grads = torch.tensor(ds["gradients"][:])
    assert not torch.isnan(grads).any()
